from ogusa import estimate_beta_j, bequest_transmission, demographics
from ogusa import macro_params, transfer_distribution, income
from ogusa import get_micro_data, psid_data_setup
import os
import numpy as np
from ogcore import txfunc
from ogcore.utils import safe_read_pickle, mkdirs
import pkg_resources


class Calibration:
    """OG-USA calibration class"""

    def __init__(
        self,
        p,
        estimate_tax_functions=False,
        estimate_beta=False,
        estimate_chi_n=False,
        tax_func_path=None,
        iit_reform={},
        guid="",
        data="cps",
        client=None,
        num_workers=1,
    ):
        self.estimate_tax_functions = estimate_tax_functions
        self.estimate_beta = estimate_beta
        self.estimate_chi_n = estimate_chi_n
        if estimate_tax_functions:
            if tax_func_path is not None:
                run_micro = False
            else:
                run_micro = True
            self.tax_function_params = self.get_tax_function_parameters(
                p,
                iit_reform,
                guid,
                data,
                client,
                num_workers,
                run_micro=run_micro,
                tax_func_path=tax_func_path,
            )
        if estimate_beta:
            self.beta_j = estimate_beta_j.beta_estimate(self)
        # if estimate_chi_n:
        #     chi_n = self.get_chi_n()

        # Macro estimation
        self.macro_params = macro_params.get_macro_params()

        # eta estimation
        self.eta = transfer_distribution.get_transfer_matrix(p.J, p.lambdas)

        # zeta estimation
        self.zeta = bequest_transmission.get_bequest_matrix(p.J, p.lambdas)

        # demographics
        self.demographic_params = demographics.get_pop_objs(
            p.E, p.S, p.T, 1, 100, p.start_year - 1, p.start_year
        )
        # demographics for 80 period lives (needed for getting e below)
        demog80 = demographics.get_pop_objs(
            20, 80, p.T, 1, 100, p.start_year - 1, p.start_year
        )

        # earnings profiles
        self.e = income.get_e_interp(
            p.S,
            self.demographic_params["omega_SS"],
            demog80["omega_SS"],
            p.lambdas,
            plot=False,
        )

    # Tax Functions
    def get_tax_function_parameters(
        self,
        p,
        iit_reform={},
        guid="",
        data="",
        client=None,
        num_workers=1,
        run_micro=False,
        tax_func_path=None,
    ):
        """
        Reads pickle file of tax function parameters or estimates the
        parameters from microsimulation model output.

        Args:
            client (Dask client object): client
            run_micro (bool): whether to estimate parameters from
                microsimulation model
            tax_func_path (string): path where find or save tax
                function parameter estimates

        Returns:
            None

        """
        # set paths if none given
        if tax_func_path is None:
            if p.baseline:
                pckl = "TxFuncEst_baseline{}.pkl".format(guid)
                tax_func_path = os.path.join(p.output_base, pckl)
                print("Using baseline tax parameters from ", tax_func_path)
            else:
                pckl = "TxFuncEst_policy{}.pkl".format(guid)
                tax_func_path = os.path.join(p.output_base, pckl)
                print(
                    "Using reform policy tax parameters from ", tax_func_path
                )
        # create directory for tax function pickles to be saved to
        mkdirs(os.path.split(tax_func_path)[0])
        # If run_micro is false, check to see if parameters file exists
        # and if it is consistent with Specifications instance
        if not run_micro:
            dict_params, run_micro = self.read_tax_func_estimate(
                p, tax_func_path
            )
            taxcalc_version = "Cached tax parameters, no taxcalc version"
        if run_micro:
            micro_data, taxcalc_version = get_micro_data.get_data(
                baseline=p.baseline,
                start_year=p.start_year,
                reform=iit_reform,
                data=data,
                path=p.output_base,
                client=client,
                num_workers=num_workers,
            )
            p.BW = len(micro_data)
            dict_params = txfunc.tax_func_estimate(  # pragma: no cover
                micro_data,
                p.BW,
                p.S,
                p.starting_age,
                p.ending_age,
                start_year=p.start_year,
                baseline=p.baseline,
                analytical_mtrs=p.analytical_mtrs,
                tax_func_type=p.tax_func_type,
                age_specific=p.age_specific,
                reform=iit_reform,
                data=data,
                client=client,
                num_workers=num_workers,
                tax_func_path=tax_func_path,
            )
        mean_income_data = dict_params["tfunc_avginc"][0]
        frac_tax_payroll = np.append(
            dict_params["tfunc_frac_tax_payroll"],
            np.ones(p.T + p.S - p.BW)
            * dict_params["tfunc_frac_tax_payroll"][-1],
        )
        # Conduct checks to be sure tax function params are consistent
        # with the model run
        params_list = ["etr", "mtrx", "mtry"]
        BW_in_tax_params = dict_params["BW"]
        start_year_in_tax_params = dict_params["start_year"]
        S_in_tax_params = len(dict_params["tfunc_etr_params_S"][0])
        # Check that start years are consistent in model and cached tax functions
        if p.start_year != start_year_in_tax_params:
            print(
                "Input Error: There is a discrepancy between the start"
                + " year of the model and that of the tax functions!!"
            )
            assert False
        # Check that S is consistent in model and cached tax functions
        # Note: even if p.age_specific = False, the arrays coming from
        # ogcore.txfunc_est should be of length S
        if p.S != S_in_tax_params:
            print(
                "Input Error: There is a discrepancy between the ages"
                + " used in the model and those in the tax functions!!"
            )
            assert False

        # Extrapolate tax function parameters for years after budget window
        # list of list: BW x S - either an array of function at that element...
        etr_params = [[None] * p.S] * p.T
        mtrx_params = [[None] * p.S] * p.T
        mtry_params = [[None] * p.S] * p.T
        for s in range(p.S):
            for t in range(p.T):
                if t < p.BW:
                    etr_params[t][s] = dict_params["tfunc_etr_params_S"][t][s]
                    mtrx_params[t][s] = dict_params["tfunc_mtrx_params_S"][t][
                        s
                    ]
                    mtry_params[t][s] = dict_params["tfunc_mtry_params_S"][t][
                        s
                    ]
                else:
                    etr_params[t][s] = dict_params["tfunc_etr_params_S"][-1][s]
                    mtrx_params[t][s] = dict_params["tfunc_mtrx_params_S"][-1][
                        s
                    ]
                    mtry_params[t][s] = dict_params["tfunc_mtry_params_S"][-1][
                        s
                    ]

        if p.constant_rates:
            print("Using constant rates!")
            # Make all tax rates equal the average
            p.tax_func_type = "linear"
            etr_params = [[None] * p.S] * p.T
            mtrx_params = [[None] * p.S] * p.T
            mtry_params = [[None] * p.S] * p.T
            for s in range(p.S):
                for t in range(p.T):
                    if t < p.BW:
                        etr_params[t][s] = dict_params["tfunc_avg_etr"][t]
                        mtrx_params[t][s] = dict_params["tfunc_avg_mtrx"][t]
                        mtry_params[t][s] = dict_params["tfunc_avg_mtry"][t]
                    else:
                        etr_params[t][s] = dict_params["tfunc_avg_etr"][-1]
                        mtrx_params[t][s] = dict_params["tfunc_avg_mtrx"][-1]
                        mtry_params[t][s] = dict_params["tfunc_avg_mtry"][-1]
        if p.zero_taxes:
            print("Zero taxes!")
            etr_params = [[0] * p.S] * p.T
            mtrx_params = [[0] * p.S] * p.T
            mtry_params = [[0] * p.S] * p.T
        tax_param_dict = {
            "etr_params": etr_params,
            "mtrx_params": mtrx_params,
            "mtry_params": mtry_params,
            "taxcalc_version": taxcalc_version,
            "mean_income_data": mean_income_data,
            "frac_tax_payroll": frac_tax_payroll,
        }

        return tax_param_dict

    def read_tax_func_estimate(self, p, tax_func_path):
        """
        This function reads in tax function parameters from pickle
        files.

        Args:
            tax_func_path (str): path to pickle with tax function
                parameter estimates

        Returns:
            dict_params (dict): dictionary containing arrays of tax
                function parameters
            run_micro (bool): whether to estimate tax function parameters

        """
        flag = 0
        if os.path.exists(tax_func_path):
            print("Tax Function Path Exists")
            dict_params = safe_read_pickle(tax_func_path)
            # check to see if tax_functions compatible
            try:
                if p.start_year != dict_params["start_year"]:
                    print(
                        "Model start year not consistent with tax "
                        + "function parameter estimates"
                    )
                    flag = 1
            except KeyError:
                pass
            try:
                p.BW = dict_params["BW"]  # QUICK FIX
                if p.BW != dict_params["BW"]:
                    print(
                        "Model budget window length is "
                        + str(p.BW)
                        + " but the tax function parameter "
                        + "estimates have a budget window length of "
                        + str(dict_params["BW"])
                    )
                    flag = 1
            except KeyError:
                pass
            try:
                if p.tax_func_type != dict_params["tax_func_type"]:
                    print(
                        "Model tax function type is not "
                        + "consistent with tax function parameter "
                        + "estimates"
                    )
                    flag = 1
            except KeyError:
                pass
            if flag >= 1:
                raise RuntimeError(
                    "Tax function parameter estimates at given path"
                    + " are not consistent with model parameters"
                    + " specified."
                )
        else:
            flag = 1
            print(
                "Tax function parameter estimates do not exist at"
                + " given path. Running new estimation."
            )
        if flag >= 1:
            dict_params = None
            run_micro = True
        else:
            run_micro = False

        return dict_params, run_micro

    # method to return all newly calibrated parameters in a dictionary
    def get_dict(self):
        dict = {}
        if self.estimate_tax_functions:
            dict.update(self.tax_function_params)
        if self.estimate_beta:
            dict["beta_annual"] = self.beta
        if self.estimate_chi_n:
            dict["chi_n"] = self.chi_n
        dict["eta"] = self.eta
        dict["zeta"] = self.zeta
        dict.update(self.macro_params)
        dict["e"] = self.e
        dict.update(self.demographic_params)

        return dict



import os
import numpy as np
import pandas as pd
from ogcore import utils

CUR_PATH = os.path.split(os.path.abspath(__file__))[0]


def get_wealth_data(
    scf_yrs_list=[2019, 2016, 2013, 2010, 2007], web=True, directory=None
):
    """
    Reads wealth data from the 2007, 2010, 2013, 2016, and 2019 Survey of
    Consumer Finances (SCF) files.

    Args:
        scf_yrs_list (list): list of SCF years to import. Currently the
            largest set of years that will work is
            [2019, 2016, 2013, 2010, 2007]
        web (Boolean): =True if function retrieves data from internet
        directory (string or None): local directory location if data are
            stored on local drive, not use internet (web=False)


    Returns:
        df_scf (Pandas DataFrame): pooled cross-sectional data from SCFs

    """
    # Hard code cpi list for given years. Index values are annual average index
    # values from monthly FRED Consumer Price Index for All Urban Consumers:
    # All Items Less Food and Energy in U.S. City Average (CPILFESL,
    # https://fred.stlouisfed.org/series/CPILFESL). Base year is 1982-1984=100.
    # Values are [263.209, 247.585, 233.810, 221.336, 210.725].
    # We then reset the base year to 2019 by dividing each annual average by
    # the 2019 annual average and multiply by 100. Base year is 2019=100
    cpi_dict = {
        "cpi2019": 100.000,
        "cpi2016": 94.06403464,
        "cpi2013": 88.83067929,
        "cpi2010": 84.09125952,
        "cpi2007": 80.05995867,
    }
    if web:
        # Throw an error if the machine is not connected to the internet
        if utils.not_connected():
            err_msg = (
                "SCF DATA ERROR: The local machine is not "
                + "connected to the internet and web=True was "
                + "selected."
            )
            raise RuntimeError(err_msg)

        file_urls = []
        for yr in scf_yrs_list:
            zipfilename = (
                "https://www.federalreserve.gov/econres/"
                + "files/scfp"
                + str(yr)
                + "s.zip"
            )
            file_urls.append(zipfilename)

        file_paths = utils.fetch_files_from_web(file_urls)

    if not web and directory is None:
        # Thow an error if web=False no source of files is given
        err_msg = (
            "SCF DATA ERROR: No local directory was "
            + "specified as the source for the data."
        )
        raise ValueError(err_msg)

    elif not web and directory is not None:
        file_paths = []
        full_directory = os.path.expanduser(directory)
        filename_list = []
        for yr in scf_yrs_list:
            filename = "rscfp" + str(yr) + ".dta"
            filename_list.append(filename)

        for name in filename_list:
            file_paths.append(os.path.join(full_directory, name))
        # Check to make sure the necessary files are present in the
        # local directory
        err_msg = (
            "hrs_by_age() ERROR: The file %s was not found in "
            + "the directory %s"
        )
        for path in file_paths:
            if not os.path.isfile(path):
                raise ValueError(err_msg % (path, full_directory))

    # read in raw SCF data to calculate moments
    scf_dict = {}
    for filename, year in zip(file_paths, scf_yrs_list):
        df_yr = pd.read_stata(filename, columns=["networth", "wgt"])
        # Add inflation adjusted net worth
        cpi = cpi_dict["cpi" + str(year)]
        df_yr["networth_infadj"] = df_yr["networth"] * cpi
        scf_dict[str(year)] = df_yr

    df_scf = scf_dict[str(scf_yrs_list[0])]
    num_yrs = len(scf_yrs_list)
    if num_yrs >= 2:
        for year in scf_yrs_list[1:]:
            df_scf = pd.concat(
                [df_scf, scf_dict[str(year)]], ignore_index=True
            )

    return df_scf


def compute_wealth_moments(scf, bin_weights):
    """
    This function computes moments (wealth shares, Gini coefficient,
    var[ln(wealth)]) from the distribution of wealth using SCF data.

    Args:
        scf (Pandas DataFrame): pooled cross-sectional data from SCFs
        bin_weights (Numpy Array) = ability weights
        J (int) = number of ability groups

    Returns:
        wealth_moments (Numpy Array): array of wealth moments

    """
    # calculate percentile shares (percentiles based on lambdas input)
    scf.sort_values(by="networth_infadj", ascending=True, inplace=True)
    scf["weight_networth"] = scf["wgt"] * scf["networth_infadj"]
    total_weight_wealth = scf.weight_networth.sum()
    cumsum = scf.wgt.cumsum()
    J = bin_weights.shape[0]
    wealth = np.zeros((J,))
    cum_weights = bin_weights.cumsum()
    for i in range(J):
        # Get number of individuals at top of percentile bin
        cutoff = scf.wgt.sum() * cum_weights[i]
        wealth[i] = (
            scf.weight_networth[cumsum < cutoff].sum()
        ) / total_weight_wealth

    wealth_share = np.zeros(J)
    wealth_share[0] = wealth[0]
    wealth_share[1:] = wealth[1:] - wealth[0:-1]

    # compute gini coeff
    scf.sort_values(by="networth_infadj", ascending=True, inplace=True)
    p = (scf.wgt.cumsum() / scf.wgt.sum()).values
    nu = ((scf.wgt * scf.networth_infadj).cumsum()).values
    nu = nu / nu[-1]
    gini_coeff = (nu[1:] * p[:-1]).sum() - (nu[:-1] * p[1:]).sum()

    # compute variance in logs
    df = scf.drop(scf[scf["networth_infadj"] <= 0.0].index)
    df["ln_networth"] = np.log(df["networth_infadj"])
    df.sort_values(by="ln_networth", ascending=True, inplace=True)
    weight_mean = ((df.ln_networth * df.wgt).sum()) / (df.wgt.sum())
    var_ln_wealth = (
        (df.wgt * ((df.ln_networth - weight_mean) ** 2)).sum()
    ) * (1.0 / (df.wgt.sum() - 1))

    wealth_moments = np.append([wealth_share], [gini_coeff, var_ln_wealth])

    return wealth_moments



"""
------------------------------------------------------------------------
This program extracts tax rate and income data from the microsimulation
model (Tax-Calculator).
------------------------------------------------------------------------
"""
from taxcalc import Records, Calculator, Policy
from pandas import DataFrame
from dask import delayed, compute
import dask.multiprocessing
import numpy as np
import os
import pickle
import pkg_resources
from ogcore import utils
from ogusa.constants import DEFAULT_START_YEAR, TC_LAST_YEAR, PUF_START_YEAR

CUR_PATH = os.path.split(os.path.abspath(__file__))[0]


def get_calculator(
    baseline,
    calculator_start_year,
    reform=None,
    data=None,
    gfactors=None,
    weights=None,
    records_start_year=PUF_START_YEAR,
):
    """
    This function creates the tax calculator object with the policy
    specified in reform and the data specified with the data kwarg.

    Args:
        baseline (boolean): True if baseline tax policy
        calculator_start_year (int): first year of budget window
        reform (dictionary): IIT policy reform parameters, None if
            baseline
        data (DataFrame or str): DataFrame or path to datafile for
            Records object
        gfactors (Tax-Calculator GrowthFactors object): growth factors
            to use to extrapolate data over budget window
        weights (DataFrame): weights for Records object
        records_start_year (int): the start year for the data and
            weights dfs (default is set to the PUF start year as defined
            in the Tax-Calculator project)

    Returns:
        calc1 (Tax-Calculator Calculator object): Calculator object with
            current_year equal to calculator_start_year

    """
    # create a calculator
    policy1 = Policy()
    if data is not None and "cps" in data:
        records1 = Records.cps_constructor()
        # impute short and long term capital gains if using CPS data
        # in 2012 SOI data 6.587% of CG as short-term gains
        records1.p22250 = 0.06587 * records1.e01100
        records1.p23250 = (1 - 0.06587) * records1.e01100
        # set total capital gains to zero
        records1.e01100 = np.zeros(records1.e01100.shape[0])
    elif data is not None:  # pragma: no cover
        records1 = Records(
            data=data,
            gfactors=gfactors,
            weights=weights,
            start_year=records_start_year,
        )  # pragma: no cover
    else:  # pragma: no cover
        records1 = Records()  # pragma: no cover

    if baseline:
        if not reform:
            print("Running current law policy baseline")
        else:
            print("Baseline policy is: ", reform)
    else:
        if not reform:
            print("Running with current law as reform")
        else:
            print("Reform policy is: ", reform)
            print("TYPE", type(reform))
    policy1.implement_reform(reform)

    # the default set up increments year to 2013
    calc1 = Calculator(records=records1, policy=policy1)

    # Check that start_year is appropriate
    if calculator_start_year > TC_LAST_YEAR:
        raise RuntimeError("Start year is beyond data extrapolation.")

    return calc1


def get_data(
    baseline=False,
    start_year=DEFAULT_START_YEAR,
    reform={},
    data=None,
    path=CUR_PATH,
    client=None,
    num_workers=1,
):
    """
    This function creates dataframes of micro data with marginal tax
    rates and information to compute effective tax rates from the
    Tax-Calculator output.  The resulting dictionary of dataframes is
    returned and saved to disk in a pickle file.

    Args:
        baseline (boolean): True if baseline tax policy
        calculator_start_year (int): first year of budget window
        reform (dictionary): IIT policy reform parameters, None if
            baseline
        data (DataFrame or str): DataFrame or path to datafile for
            Records object
        path (str): path to save microdata files to
        client (Dask Client object): client for Dask multiprocessing
        num_workers (int): number of workers to use for Dask
            multiprocessing

    Returns:
        micro_data_dict (dict): dict of Pandas Dataframe, one for each
            year from start_year to the maximum year Tax-Calculator can
            analyze
        taxcalc_version (str): version of Tax-Calculator used

    """
    # Compute MTRs and taxes or each year, but not beyond TC_LAST_YEAR
    lazy_values = []
    for year in range(start_year, TC_LAST_YEAR + 1):
        lazy_values.append(
            delayed(taxcalc_advance)(baseline, start_year, reform, data, year)
        )
    if client:  # pragma: no cover
        futures = client.compute(lazy_values, num_workers=num_workers)
        results = client.gather(futures)
    else:
        results = results = compute(
            *lazy_values,
            scheduler=dask.multiprocessing.get,
            num_workers=num_workers,
        )

    # dictionary of data frames to return
    micro_data_dict = {}
    for i, result in enumerate(results):
        year = start_year + i
        micro_data_dict[str(year)] = DataFrame(result)

    if baseline:
        pkl_path = os.path.join(path, "micro_data_baseline.pkl")
    else:
        pkl_path = os.path.join(path, "micro_data_policy.pkl")
    utils.mkdirs(path)
    with open(pkl_path, "wb") as f:
        pickle.dump(micro_data_dict, f)

    # Do some garbage collection
    del results

    # Pull Tax-Calc version for reference
    taxcalc_version = pkg_resources.get_distribution("taxcalc").version

    return micro_data_dict, taxcalc_version


def taxcalc_advance(baseline, start_year, reform, data, year):
    """
    This function advances the year used in Tax-Calculator, compute
    taxes and rates, and save the results to a dictionary.

    Args:
        calc1 (Tax-Calculator Calculator object): TC calculator
        year (int): year to begin advancing from

    Returns:
        tax_dict (dict): a dictionary of microdata with marginal tax
            rates and other information computed in TC
    """
    calc1 = get_calculator(
        baseline=baseline,
        calculator_start_year=start_year,
        reform=reform,
        data=data,
    )
    calc1.advance_to_year(year)
    calc1.calc_all()
    print("Year: ", str(calc1.current_year))

    # define market income - taking expanded_income and excluding gov't
    # transfer benefits found in the Tax-Calculator expanded income
    market_income = calc1.array("expanded_income") - calc1.array(
        "benefit_value_total"
    )

    # Compute mtr on capital income
    mtr_combined_capinc = cap_inc_mtr(calc1)

    # Compute weighted avg mtr for labor income
    # Note the index [2] in the mtr results means that we are pulling
    # the combined mtr from the IIT + FICA taxes
    mtr_combined_labinc = (
        calc1.mtr("e00200p")[2] * np.abs(calc1.array("e00200"))
        + calc1.mtr("e00900p")[2] * np.abs(calc1.array("sey"))
    ) / (np.abs(calc1.array("sey")) + np.abs(calc1.array("e00200")))

    # Put MTRs, income, tax liability, and other variables in dict
    length = len(calc1.array("s006"))
    tax_dict = {
        "mtr_labinc": mtr_combined_labinc,
        "mtr_capinc": mtr_combined_capinc,
        "age": calc1.array("age_head"),
        "total_labinc": calc1.array("sey") + calc1.array("e00200"),
        "total_capinc": (
            market_income - calc1.array("sey") + calc1.array("e00200")
        ),
        "market_income": market_income,
        "total_tax_liab": calc1.array("combined"),
        "payroll_tax_liab": calc1.array("payrolltax"),
        "etr": (
            (calc1.array("combined") - calc1.array("ubi")) / market_income
        ),
        "year": calc1.current_year * np.ones(length),
        "weight": calc1.array("s006"),
    }

    # garbage collection
    del calc1

    return tax_dict


def cap_inc_mtr(calc1):  # pragma: no cover
    """
    This function computes the marginal tax rate on capital income,
    which is calculated as a weighted average of the marginal tax rates
    on different sources of capital income.

    Args:
        calc1 (Tax-Calculator Calculator object): TC calculator

    Returns:
        mtr_combined_capinc (Numpy array): array with marginal tax rates
            for each observation in the TC Records object

    """
    # Note: PUF does not have variable for non-taxable IRA distributions
    # Exclude Sch E income (e02000) from this list since we'll compute
    # MTRs for this income in two parts - one for overall Sch C and one
    # for S Corp and Partnerhsip income (e26270) (note that TaxCalc
    # doesn't allow for an MTR on rents and royalties alone)
    # e00300 = interest income
    # e00400 = nontaxable interest income
    # e00600 = ordinary dividend income
    # e00650 = qualified dividend income
    # e01400 = taxable IRA distributions
    # e01700 = pension and annuity income
    # p22250 = short term cap gain/loss
    # p23250 = long term cap gain/loss
    # e26270 = partnership and s corp income/loss
    # e02000 = Sch E income (includes e26270)
    capital_income_sources = (
        "e00300",
        "e00400",
        "e00600",
        "e00650",
        "e01400",
        "e01700",
        "p22250",
        "p23250",
        "e26270",
    )
    rent_royalty_inc = np.abs(calc1.array("e02000") - calc1.array("e26270"))
    # assign overall Sch E mtr to rent and royalities since TC can't do
    # this component separately
    rent_royalty_mtr = calc1.mtr("e02000")[2]
    # calculating MTRs separately - can skip items with zero tax
    all_mtrs = {
        income_source: calc1.mtr(income_source)
        for income_source in capital_income_sources
    }
    # Get each column of income sources, to include non-taxable income
    record_columns = [calc1.array(x) for x in capital_income_sources]
    # Compute weighted average of all those MTRs
    # first find total capital income
    total_cap_inc = sum(map(abs, record_columns)) + rent_royalty_inc
    # Note that all_mtrs gives fica (0), iit (1), and combined (2) mtrs
    # We'll use the combined - hence all_mtrs[source][2]
    capital_mtr = [
        abs(col) * all_mtrs[source][2]
        for col, source in zip(record_columns, capital_income_sources)
    ]
    mtr_combined_capinc = np.zeros_like(total_cap_inc)
    mtr_combined_capinc[total_cap_inc != 0] = (
        sum(capital_mtr + rent_royalty_mtr * rent_royalty_inc)[
            total_cap_inc != 0
        ]
        / total_cap_inc[total_cap_inc != 0]
    )
    mtr_combined_capinc[total_cap_inc == 0] = all_mtrs["e00300"][2][
        total_cap_inc == 0
    ]
    return mtr_combined_capinc



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
from ogusa.utils import MVKDE

CURDIR = os.path.split(os.path.abspath(__file__))[0]


def get_bequest_matrix(
    J=7, lambdas=np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.01]), graphs=True
):
    """
    Returns S x J matrix representing the fraction of aggregate
    bequests that go to each household by age and lifetime income group.

    """
    # Create directory if output directory does not already exist
    CURDIR = os.path.split(os.path.abspath(__file__))[0]
    output_fldr = "csv_output_files"
    output_dir = os.path.join(CURDIR, "..", "data", output_fldr)
    if not os.access(output_dir, os.F_OK):
        os.makedirs(output_dir)
    image_fldr = "images"
    image_dir = os.path.join(CURDIR, "..", "data", image_fldr)
    if not os.access(image_dir, os.F_OK):
        os.makedirs(image_dir)

    # Define a lambda function to compute the weighted mean:
    # wm = lambda x: np.average(
    #     x, weights=df.loc[x.index, "fam_smpl_wgt_core"])

    # Read in dataframe of PSID data
    # df = ogcore.utils.safe_read_pickle(
    #     os.path.join(CURDIR, "data", "PSID", "psid_lifetime_income.pkl")
    # )
    df = pd.read_csv(
        os.path.join(CURDIR, "..", "data", "PSID", "psid_lifetime_income.csv")
    )

    # Do some tabs with data file...
    # 'net_wealth', 'inheritance', 'value_inheritance_1st',
    # 'value_inheritance_2nd', 'value_inheritance_3rd's
    # inheritance available from 1988 onwards...
    # Note that the resulting distribution is very different from what
    # Rick has found with the SCF

    df["sum_inherit"] = (
        df["value_inheritance_1st"]
        + df["value_inheritance_2nd"]
        + df["value_inheritance_3rd"]
    )
    # print(df[['sum_inherit', 'inheritance']].describe())

    if graphs:
        # Total inheritances by year
        df.groupby("year_data").mean(numeric_only=True).plot(y="inheritance")
        plt.savefig(os.path.join(image_dir, "inheritance_year.png"))
        df.groupby("year_data").mean(numeric_only=True).plot(y="sum_inherit")
        plt.savefig(os.path.join(image_dir, "sum_inherit_year.png"))
        # not that summing up inheritances gives a much larger value than
        # taking the inheritance variable

        # Fraction of inheritances in a year by age
        # line plot
        df[df["year_data"] >= 1988].groupby("age").mean(
            numeric_only=True
        ).plot(y="net_wealth")
        plt.savefig(os.path.join(image_dir, "net_wealth_age.png"))
        df[df["year_data"] >= 1988].groupby("age").mean(
            numeric_only=True
        ).plot(y="inheritance")
        plt.savefig(os.path.join(image_dir, "inheritance_age.png"))

        # Inheritances by lifetime income group
        # bar plot
        df[df["year_data"] >= 1988].groupby("li_group").mean(
            numeric_only=True
        ).plot.bar(y="net_wealth")
        plt.savefig(os.path.join(image_dir, "net_wealth_li.png"))
        df[df["year_data"] >= 1988].groupby("li_group").mean(
            numeric_only=True
        ).plot.bar(y="inheritance")
        plt.savefig(os.path.join(image_dir, "inheritance_li.png"))

        # lifecycle plots with line for each ability type
        pd.pivot_table(
            df[df["year_data"] >= 1988],
            values="net_wealth",
            index="age",
            columns="li_group",
            aggfunc="mean",
        ).plot(legend=True)
        plt.savefig(os.path.join(image_dir, "net_wealth_age_li.png"))
        pd.pivot_table(
            df[df["year_data"] >= 1988],
            values="inheritance",
            index="age",
            columns="li_group",
            aggfunc="mean",
        ).plot(legend=True)
        plt.savefig(os.path.join(image_dir, "inheritance_age_li.png"))

    # Matrix Fraction of inheritances in a year by age and lifetime_inc
    inheritance_matrix = pd.pivot_table(
        df[df["year_data"] >= 1988],
        values="inheritance",
        index="age",
        columns="li_group",
        aggfunc="sum",
    )
    # replace NaN with zero
    inheritance_matrix.fillna(value=0, inplace=True)
    inheritance_matrix = inheritance_matrix / inheritance_matrix.sum().sum()
    # inheritance_matrix.to_csv(os.path.join(
    #     output_dir, 'bequest_matrix.csv'))

    # estimate kernel density of bequests
    kde_matrix = MVKDE(
        80,
        7,
        inheritance_matrix.to_numpy(),
        filename=os.path.join(image_dir, "inheritance_kde.png"),
        plot=graphs,
        bandwidth=0.5,
    )

    if (J == 10) and np.array_equal(
        np.squeeze(lambdas[:6]), np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09])
    ):
        kde_matrix_new = np.zeros((80, J))
        kde_matrix_new[:, :6] = kde_matrix[:, :6]
        kde_matrix_new[:, 6:] = (
            kde_matrix[:, 6:].sum(axis=1).reshape(80, 1)
            * np.tile(np.reshape(lambdas[6:], (1, 4)), (80, 1))
            / lambdas[6:].sum()
        )
        kde_matrix = kde_matrix_new

    np.savetxt(
        os.path.join(output_dir, "bequest_matrix_kde.csv"),
        kde_matrix,
        delimiter=",",
    )

    return kde_matrix



import pandas as pd
import numpy as np
from scipy.stats import kde
import matplotlib.pyplot as plt
import requests
import urllib3
import ssl


def read_cbo_forecast():
    """
    This function reads the CBO Long-Term Budget Projections document
    from https://www.cbo.gov/about/products/budget-economic-data#1
    and then formats the relevant data for use with OG-Core
    """
    CBO_LT_URL = (
        "https://www.cbo.gov/system/files/2020-09/51119-2020-09-ltbo_0.xlsx"
    )
    # Read in data
    df = pd.read_excel(
        CBO_LT_URL, sheet_name="3. Economic Vars", skiprows=7, nrows=45
    )
    df.drop(columns=["Unnamed: 3", "Unnamed: 4"], inplace=True)
    df[
        ~(
            (pd.isnull(df["Unnamed: 0"]))
            & (pd.isnull(df["Unnamed: 1"]))
            & (pd.isnull(df["Unnamed: 2"]))
        )
    ]
    df.fillna(value="", inplace=True)
    df["full_var_name"] = (
        df["Unnamed: 0"] + df["Unnamed: 1"] + df["Unnamed: 2"]
    )
    CBO_VAR_NAMES = {
        "Real GDP (Billions of 2019 dollars) ": "Y",
        "On 10-year Treasury notes and the OASDI trust funds": "r",
        "Growth of Real Earnings per Worker": "w_growth",
        "Growth of Total Hours Worked": "L_growth",
        "Hours of All Persons (Nonfarm Business Sector)": "L",
        "Personal Consumption Expenditures": "C",
        "Gross Private Domestic Investment": "I_total",
        "Government Consumption Expenditures and Gross Investment": "G",
        "Old-Age and Survivors Insurance": "agg_pension_outlays",
        "Individual income taxes": "iit_revenue",
        "Payroll taxes": "payroll_tax_revenue",
        "Corporate income taxes": "business_tax_revenue",
        "Wages and Salaries": "wL",
    }
    df["var_name"] = df["full_var_name"].replace(CBO_VAR_NAMES)
    # keep just variables of interest
    df.drop(
        columns=["Unnamed: 0", "Unnamed: 1", "Unnamed: 2", "full_var_name"],
        inplace=True,
    )
    df = df[df["var_name"].isin(CBO_VAR_NAMES.values())]
    # Keep just real interest rate (not nominal)
    # Note that real interest rate comes first in table
    df.drop_duplicates(subset="var_name", inplace=True)
    # reshape so that variable names down column
    df = pd.melt(
        df, id_vars="var_name", value_vars=[i for i in range(1990, 2051)]
    )
    df = df.pivot(index="variable", columns="var_name", values="value")
    df.reset_index(inplace=True)
    df.rename(columns={"variable": "year"}, inplace=True)
    # add debt forcast
    df_fiscal = pd.read_excel(
        CBO_LT_URL,
        sheet_name="1. Summary Extended Baseline",
        skiprows=9,
        nrows=32,
    )
    df_fiscal = df_fiscal[
        ["Fiscal Year", "Revenues", "Federal Debt Held by the Public"]
    ]
    df_lt = df.merge(
        df_fiscal, left_on="year", right_on="Fiscal Year", how="left"
    )
    df_lt.rename(
        columns={"Federal Debt Held by the Public": "D/Y"}, inplace=True
    )
    df_lt["D"] = df_lt["Y"] * df_lt["D/Y"]

    CBO_10yr_budget_URL = (
        "https://www.cbo.gov/system/files/2021-02/51118-2021-02-11-"
        + "budgetprojections.xlsx"
    )
    df = pd.read_excel(
        CBO_10yr_budget_URL, sheet_name="Table 1-1", skiprows=8, nrows=7
    )
    df.rename(
        columns={"Unnamed: 0": "variable", "Actual, \n2020": 2020},
        inplace=True,
    )
    df.drop(columns=["2026.1", "2031.1"], inplace=True)
    df1 = df[~((pd.isnull(df.variable)) | (df.variable == "Other"))]

    df = pd.read_excel(
        CBO_10yr_budget_URL, sheet_name="Table 1-3", skiprows=9, nrows=22
    )
    df.rename(columns={"Unnamed: 0": "variable"}, inplace=True)
    df.drop(columns=["2026.1", "2031.1"], inplace=True)
    df.drop_duplicates(subset="variable", keep="last", inplace=True)
    df2 = df[~pd.isnull(df.variable)]

    CBO_10yr_macro_URL = (
        "https://www.cbo.gov/system/files/2021-02/51135-2021-02-"
        + "economicprojections.xlsx"
    )
    df = pd.read_excel(
        CBO_10yr_macro_URL,
        sheet_name="2. Calendar Year",
        skiprows=6,
        nrows=131,
    )
    df.rename(columns={"Unnamed: 1": "variable"}, inplace=True)
    df.drop(columns=["Unnamed: 0", "Unnamed: 2", "Units"], inplace=True)
    # Note that real values come second (after nominal values)
    df.drop_duplicates(subset="variable", keep="last", inplace=True)
    df3 = df[~pd.isnull(df.variable)]
    df_st = pd.concat([df1, df2, df3], sort=False, ignore_index=True)
    # df_st = df1.append(df2, sort=False, ignore_index=True).append(
    #     df3, sort=False, ignore_index=True
    # )
    df_st["var_name"] = df_st["variable"].replace(CBO_VAR_NAMES)
    df_st = df_st[~pd.isnull(df_st.var_name)]
    df_st.drop(columns=["variable"], inplace=True)
    # reshape so each row a year
    df_st = pd.melt(
        df_st, id_vars="var_name", value_vars=[i for i in range(2017, 2031)]
    )
    df_st = df_st.pivot(
        index="variable", columns="var_name", values="value"
    ).reset_index()
    df_st.rename(columns={"variable": "year"}, inplace=True)

    # merge with long term data
    df_cbo = df_lt.merge(
        df_st, how="outer", on="year", suffixes=("_lt", "_st")
    )
    # replace * with 0
    df_cbo.replace(to_replace="*", value=0.0, inplace=True)

    return df_cbo


# Will need to do some smoothing with a KDE when estimate the matrix...
def MVKDE(
    S,
    J,
    proportion_matrix,
    zaxis_label="Received proportion of total transfers",
    filename=None,
    plot=False,
    bandwidth=0.25,
):
    """
    Generates a Multivariate Kernel Density Estimator and returns a
    matrix representing a probability distribution according to given
    age categories, and ability type categories.

    Args:
        S (scalar): the number of age groups in the model
        J (scalar): the number of ability type groups in the model.
        proportion_matrix (Numpy array): SxJ shaped array that
            represents the proportions of the total going to each
            (s,j) combination
        filename (str): the file name  to save image to
        plot (bool): whether or not to save a plot of the probability
            distribution generated by the kde or the proportion matrix
        bandwidth (scalar):  used in the smoothing of the kernel. Higher
            bandwidth creates a smoother kernel.

    Returns:
        estimator_scaled (Numpy array): SxJ shaped array that
            that represents the smoothed distribution of proportions
            going to each (s,j)

    """
    proportion_matrix_income = np.sum(proportion_matrix, axis=0)
    proportion_matrix_age = np.sum(proportion_matrix, axis=1)
    age_probs = np.random.multinomial(70000, proportion_matrix_age)
    income_probs = np.random.multinomial(70000, proportion_matrix_income)
    age_frequency = np.array([])
    income_frequency = np.array([])
    age_mesh = complex(str(S) + "j")
    income_mesh = complex(str(J) + "j")
    j = 18
    """creating a distribution of age values"""
    for i in age_probs:
        listit = np.ones(i)
        listit *= j
        age_frequency = np.append(age_frequency, listit)
        j += 1

    k = 1
    """creating a distribution of ability type values"""
    for i in income_probs:
        listit2 = np.ones(i)
        listit2 *= k
        income_frequency = np.append(income_frequency, listit2)
        k += 1

    freq_mat = np.vstack((age_frequency, income_frequency)).T
    density = kde.gaussian_kde(freq_mat.T, bw_method=bandwidth)
    age_min, income_min = freq_mat.min(axis=0)
    age_max, income_max = freq_mat.max(axis=0)
    agei, incomei = np.mgrid[
        age_min:age_max:age_mesh, income_min:income_max:income_mesh
    ]
    coords = np.vstack([item.ravel() for item in [agei, incomei]])
    estimator = density(coords).reshape(agei.shape)
    estimator_scaled = estimator / float(np.sum(estimator))
    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")
        ax.plot_surface(agei, incomei, estimator_scaled, rstride=5)
        ax.set_xlabel("Age")
        ax.set_ylabel("Ability Types")
        ax.set_zlabel(zaxis_label)
        plt.savefig(filename)
    return estimator_scaled


class CustomHttpAdapter(requests.adapters.HTTPAdapter):
    """
    The UN Data Portal server doesn't support "RFC 5746 secure renegotiation". This causes and error when the client is using OpenSSL 3, which enforces that standard by default.
    The fix is to create a custom SSL context that allows for legacy connections. This defines a function get_legacy_session() that should be used instead of requests().
    """

    # "Transport adapter" that allows us to use custom ssl_context.
    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_context=self.ssl_context,
        )


def get_legacy_session():
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT  #in Python 3.12 you will be able to switch from 0x4 to ssl.OP_LEGACY_SERVER_CONNECT.
    session = requests.session()
    session.mount("https://", CustomHttpAdapter(ctx))
    return session



"""
This script uses simulated method of moments estimator to estimate the
beta_j parameters for OG-USA.
"""


import numpy as np
import os
import scipy.optimize as opt
from ogcore.parameters import Specifications
from ogusa import wealth
from ogcore import SS
from ogcore.utils import Inequality


# Don't print output along the way of SS solution
SS.VERBOSE = False

CUR_PATH = os.path.split(os.path.abspath(__file__))[0]


def beta_estimate(
    beta_initial_guesses, og_spec={}, two_step=False, client=None
):
    """
    This function estimates the beta_j parameters using a simulated
    method of moments estimator that targets moments of the wealth
    distribution.

    Args:
        beta_initial_guesses (array-like): array of initial guesses for the
            beta_j parameters
        og_spec (dict): any updates to default model parameters
        two_step (boolean): whether to use a two-step estimator
        client (Dask Client object): dask client for multiprocessing

    Returns:
        beta_hat (array-like): estimates of the beta_j
        beta_se (array-like): standard errors on the beta_j estimates

    """

    # initialize parameters object
    tax_func_path = os.path.join(
        CUR_PATH,
        "..",
        "data",
        "tax_functions",
        "TxFuncEst_baseline_PUF.pkl",
    )
    p = Specifications(baseline=True)
    p.update_specifications(og_spec)
    p.get_tax_function_parameters(client, False, tax_func_path)

    # Compute wealth moments from the data
    scf = wealth.get_wealth_data(scf_yrs_list=[2019], web=True, directory=None)
    data_moments = wealth.compute_wealth_moments(scf, p.lambdas)

    # Get weighting matrix
    W = compute_weighting_matrix(p, optimal_weight=False)

    # call minimizer
    # set bounds on beta estimates (need to be between 0 and 1)
    bnds = np.tile(np.array([1e-12, 1]), (p.J, 1))  # Need (1e-12, 1) J times
    # pack arguments in a tuple
    min_args = (data_moments, W, p, client)
    # NOTE: may want to try some global optimization routing like
    # simulated annealing (aka basin hopping) or differential
    # evolution
    est_output = opt.minimize(
        minstat,
        beta_initial_guesses,
        args=(min_args),
        method="L-BFGS-B",
        bounds=bnds,
        tol=1e-15,
        options={"maxfun": 1, "maxiter": 1, "maxls": 1},
    )
    beta_hat = est_output["x"]

    # calculate std errors
    K = len(data_moments)
    beta_se, VCV_params = compute_se(beta_hat, W, K, p, h=0.01, client=client)

    if two_step:
        W = VCV_params
        min_args = (data_moments, W, p, client)
        est_output = opt.minimize(
            minstat,
            beta_initial_guesses,
            args=(min_args),
            method="L-BFGS-B",
            bounds=bnds,
            tol=1e-15,
            options={"maxfun": 1, "maxiter": 1, "maxls": 1},
        )
        beta_hat = est_output["x"]
        beta_se, VCV_params = compute_se(
            beta_hat, W, K, p, h=0.01, client=client
        )

    return beta_hat, beta_se


def minstat(beta_guesses, *args):
    """
    This function generates the weighted sum of squared differences
    between the model and data moments.

    Args:
        beta_guesses (array-like): a vector of length J with the betas
        args (tuple): length 6 tuple, variables needed for minimizer

    Returns:
        distance (scalar): weighted, squared deviation between data and
            model moments

    """
    # unpack args tuple
    data_moments, W, p, client = args

    # Update beta in parameters object with beta guesses
    p.beta = beta_guesses

    # Solve model SS
    print("Baseline = ", p.baseline)
    ss_output = SS.run_SS(p, client=client)

    # Compute moments from model SS
    model_moments = calc_moments(ss_output, p)

    # distance with levels
    distance = np.dot(
        np.dot((np.array(model_moments) - np.array(data_moments)).T, W),
        np.array(model_moments) - np.array(data_moments),
    )

    print("DATA and MODEL DISTANCE: ", distance)

    return distance


def calc_moments(ss_output, p):
    """
    This function calculates moments from the SS output that correspond
    to the data moments used for estimation.

    Args:
        ss_output = dictionary, variables from SS of model
        p (OG-USA Specifications object): model parameters

    Returns:
        model_moments (array-like): Array of model moments

    """
    # Create Inequality object
    wealth_ineq = Inequality(
        ss_output["bssmat_splus1"], p.omega_SS, p.lambdas, p.S, p.J
    )

    # wealth moments
    # moments are: bottom 25% of wealth, next 25% share of wealth
    #  (25-50 pctile), next 20% share of wealth (50-70 pctile),
    #  next 10% share (70-80 pctile), next 10% share (80-90 pctile),
    #  next 9% share (90-99 pctile), top 1% share,
    # gini coefficient, variance of log wealth
    model_moments = np.array(
        [
            1 - wealth_ineq.top_share(0.75),
            wealth_ineq.top_share(0.75) - wealth_ineq.top_share(0.50),
            wealth_ineq.top_share(0.50) - wealth_ineq.top_share(0.30),
            wealth_ineq.top_share(0.30) - wealth_ineq.top_share(0.20),
            wealth_ineq.top_share(0.20) - wealth_ineq.top_share(0.10),
            wealth_ineq.top_share(0.10) - wealth_ineq.top_share(0.01),
            wealth_ineq.top_share(0.01),
            wealth_ineq.gini(),
            wealth_ineq.var_of_logs(),
        ]
    )

    return model_moments


def compute_weighting_matrix(p, optimal_weight=False):
    """
    Function to compute the weighting matrix for the GMM estimator.

    Args:
        p (OG-USA Specifications object): model parameters
        optimal_weight (boolean): whether to use an optimal
            weighting matrix or not

    Returns:
        W (Numpy array): Weighting matrix

    """
    # determine weighting matrix
    if optimal_weight:
        # This uses the inverse of the VCV matrix for the data moments
        # more precisely estimated moments get more weight
        # Reference: Gourieroux, Monfort, and Renault (1993,
        # Journal of Applied Econometrics)
        # read in SCF
        n = 1000  # number of bootstrap iterations
        scf = wealth.get_wealth_data(
            scf_yrs_list=[2019], web=True, directory=None
        )
        VCV_data_moments = wealth.VCV_moments(scf, n, p.lambdas, p.J)
        W = np.linalg.inv(VCV_data_moments)
    else:
        # Assumes use 2 more moments than there are parameters
        W = np.identity(p.J + 2)

    return W


def VCV_moments(scf, n, bin_weights, J):
    """
    Compute variance-covariance matrix for wealth moments by
    bootstrapping data.

    Args:
        scf (Pandas DataFrame): raw data from SCF
        n (int): number of bootstrap iterations to run
        bin_weights (array-like): ability weights (Jx1 array)
        J (int): number of ability groups

    Returns:
        VCV (Numpy array): variance-covariance matrix of wealth moments,
            (J+2xJ+2) array

    """
    wealth_moments_boot = np.zeros((n, J + 2))
    for i in range(n):
        sample = scf[np.random.randint(2, size=len(scf.index)).astype(bool)]
        # note that wealth moments from data are in array in same order
        # as model moments are computed in this module
        wealth_moments_boot[i, :] = wealth.compute_wealth_moments(
            sample, bin_weights
        )

    VCV = np.cov(wealth_moments_boot.T)

    return VCV


def compute_se(beta_hat, W, K, p, h=0.01, client=None):
    """
    Function to compute standard errors for the SMM estimator.

    Args:
        beta_hat (array-like): estimates of beta parameters
        W (Numpy array): weighting matrix
        K (int): number of moments
        p (OG-USA Specifications object): model parameters
        h (scalar): percentage to move parameters for numerical derivatives
        client (Dask Client object): Dask client

    Returns:
        beta_se (array-like): standard errors for beta estimates
        VCV_params (Numpy array): VCV matrix for parameter estimates

    """
    # compute numerical derivatives that will need for SE's
    model_moments_low = np.zeros((p.J, K))
    model_moments_high = np.zeros((p.J, K))
    beta_low = beta_hat
    beta_high = beta_hat
    for i in range(len(beta_hat)):
        # compute moments with downward change in param
        beta_low[i] = beta_hat[i] * (1 + h)
        p.beta = beta_low
        ss_output = ss_output = SS.run_SS(p, client=client)
        model_moments_low[i, :] = calc_moments(ss_output, p)
        # compute moments with upward change in param
        beta_high[i] = beta_hat[i] * (1 - h)
        p.beta = beta_low
        ss_output = ss_output = SS.run_SS(p, client=client)
        model_moments_high[i, :] = calc_moments(ss_output, p)

    deriv_moments = (model_moments_high - model_moments_low).T / (
        2 * h * beta_hat
    )
    VCV_params = np.linalg.inv(
        np.dot(np.dot(deriv_moments.T, W), deriv_moments)
    )
    beta_se = (np.diag(VCV_params)) ** (1 / 2)

    return beta_se, VCV_params



import numpy as np
import pandas as pd
import os
import pickle
from pandas_datareader import data as web
import datetime
from linearmodels import PanelOLS
from rpy2.robjects import r
from rpy2.robjects import pandas2ri
from ogusa.constants import PSID_NOMINAL_VARS, PSID_CONSTANT_VARS

pandas2ri.activate()
pd.options.mode.chained_assignment = "raise"

CURDIR = os.path.split(os.path.abspath(__file__))[0]


def prep_data(data="psid1968to2015.RData"):
    """
    This script takes PSID data created from psid_download.R and:
    1) Creates variables at the "tax filing unit" (equal to family
        unit in PSID since there is no info on the filing status chosen).
    2) Selects a sample of observations to work with (e.g., dropping
        very old, very low income, etc.).
    3) Computes a measure of lifetime income and places each household
        into a lifetime income percentile group

    Args:
        data (str): path to RData file with PSID data

    Returns:
        panel_li (Pandas DataFrame): household level data with lifetime
            income groups defined
    """
    # Read data from R into pandas dataframe
    r["load"](os.path.join(CURDIR, "..", "data", "PSID", data))
    raw_df = r("psid_df")

    # Create unique identifier for each household
    # note that will define a new household if head or spouse changes
    # keep only current heads
    # before 1983, head is relation.head == 1, 1983+ head is given by
    # relation.head == 10

    # Select just those in the SRC sample, which is representative of the
    # population and so will not require the use of sampling weights
    # SRC sample families have 1968 family interview numbers less than 3000
    raw_df = raw_df[raw_df["ID1968"] < 3000].copy()

    raw_df["relation.head"][
        (raw_df["year"] < 1983) & (raw_df["relation.head"] == 1)
    ] = 10
    raw_df["relation.head"][
        (raw_df["year"] < 1983) & (raw_df["relation.head"] == 2)
    ] = 20
    head_df = raw_df.loc[
        raw_df.index[
            (raw_df["relation.head"] == 10) & (raw_df["sequence"] == 1)
        ],
        :,
    ]
    head_df.rename(columns={"pid": "head_id"}, inplace=True)
    # keep legal spouse or long term partners
    spouse_df = raw_df.loc[
        raw_df.index[
            (raw_df["relation.head"] >= 20)
            & (raw_df["relation.head"] <= 22)
            & (raw_df["sequence"] == 2)
        ],
        ["pid", "ID1968", "year", "interview_number"],
    ]
    spouse_df.rename(columns={"pid": "spouse_id"}, inplace=True)
    psid_df = head_df.merge(
        spouse_df, how="left", on=["ID1968", "year", "interview_number"]
    )
    # create unique household id for combination of head and a specific spouse
    psid_df["hh_id"] = (psid_df["head_id"] * 1000000) + psid_df[
        "spouse_id"
    ].fillna(0)

    # clean up files no longer need
    del raw_df, head_df, spouse_df

    # Fix ages to increment by one (or two) between survey waves.  They do
    # not always do this because the survey may be asked as different times
    # of year
    min_age_df = psid_df.groupby("hh_id").agg(["min"])["head_age"]
    min_age_df.rename(columns={"min": "min_age"}, inplace=True)
    min_year_df = psid_df.groupby("hh_id").agg(["min"])["year"]
    min_year_df.rename(columns={"min": "min_year"}, inplace=True)
    psid_df = psid_df.merge(min_age_df, on="hh_id", how="left")
    psid_df = psid_df.merge(min_year_df, on="hh_id", how="left")
    psid_df.sort_values(by=["hh_id", "year"], inplace=True)
    psid_df["age"] = psid_df["year"] - psid_df["min_year"] + psid_df["min_age"]

    # clean up
    del min_age_df, min_year_df

    # Deflate nominal variables
    # because surveys ask about prior year
    psid_df["year_data"] = psid_df["year"] - 1

    # create spouse labor income, since not consistent variable name
    # across time
    psid_df["spouse_labor_inc"] = (
        psid_df["spouse_labor_inc_pre1993"]
        + psid_df["spouse_labor_inc_post1993"]
    )
    psid_df.loc[psid_df["year"] == 1993, "spouse_labor_inc"] = psid_df[
        "spouse_labor_inc_post1993"
    ]

    # set beginning and end dates for data
    start = datetime.datetime(1968, 1, 1)
    end = datetime.datetime.today()
    # pull series of interest using pandas_datareader
    fred_data = web.DataReader(["CPIAUCSL"], "fred", start, end)
    # Make data annual by averaging over months in year
    fred_data = fred_data.resample("A").mean()
    fred_data["year_data"] = fred_data.index.year
    psid_df2 = psid_df.merge(fred_data, how="left", on="year_data")
    psid_df = psid_df2
    cpi_2010 = fred_data.loc[datetime.datetime(2010, 12, 31), "CPIAUCSL"]

    for item in PSID_NOMINAL_VARS:
        psid_df[item] = (psid_df[item] * cpi_2010) / psid_df["CPIAUCSL"]

    # clean up
    del fred_data, psid_df2

    # Fill in  missing values with zeros
    psid_df[PSID_NOMINAL_VARS] = psid_df[PSID_NOMINAL_VARS].fillna(0)
    psid_df[["head_annual_hours", "spouse_annual_hours"]] = psid_df[
        ["head_annual_hours", "spouse_annual_hours"]
    ].fillna(0)

    # Construct family ("filing unit") level variables
    psid_df["incwage_hh"] = (
        psid_df["head_labor_inc"] + psid_df["spouse_labor_inc"]
    )
    psid_df["earninc_hh"] = (
        psid_df["incwage_hh"]
        + psid_df["head_noncorp_bus_labor_income"]
        + psid_df["spouse_noncorp_bus_labor_income"]
    )
    psid_df["businc_hh"] = (
        psid_df["head_noncorp_bus_labor_income"]
        + psid_df["spouse_noncorp_bus_labor_income"]
    )
    # note that PSID doesn't separate hours towards employed and business
    # work
    psid_df["earnhours_hh"] = (
        psid_df["head_annual_hours"] + psid_df["spouse_annual_hours"]
    )
    psid_df["wage_rate"] = psid_df["incwage_hh"] / psid_df["earnhours_hh"]
    psid_df["earn_rate"] = psid_df["earninc_hh"] / psid_df["earnhours_hh"]
    with np.errstate(divide="ignore"):
        psid_df["ln_wage_rate"] = np.log(psid_df["wage_rate"])
        psid_df["ln_earn_rate"] = np.log(psid_df["earn_rate"])
    psid_df["singlemale"] = (psid_df["head_gender"] == 1) & (
        psid_df["marital_status"] != 1
    )
    psid_df["singlefemale"] = (psid_df["head_gender"] == 2) & (
        psid_df["marital_status"] != 1
    )
    psid_df["marriedmalehead"] = (psid_df["head_gender"] == 1) & (
        psid_df["marital_status"] == 1
    )
    psid_df["marriedfemalehead"] = (psid_df["head_gender"] == 2) & (
        psid_df["marital_status"] == 1
    )
    psid_df["married"] = (psid_df["marital_status"] == 1).astype(int)

    # sample selection
    # drop very young, very old, those with very low earnings, and any
    # outliers with very high earnings, those working at least 200 hrs
    # should check to see if we want to drop any particular years... (e.g.,
    # I think some data is missing before 1970)
    psid_df.query(
        "age >= 20 & age <= 80 & incwage_hh >= 5"
        + " & wage_rate >= 5 & wage_rate <= 25000"
        + " & earnhours_hh > 200",
        inplace=True,
    )
    # Indicator for obs being from PSID not interpolated value
    # used to make drops later
    psid_df.sort_values(by=["hh_id", "year"], inplace=True)
    # psid_df[
    #     [
    #         "head_id",
    #         "spouse_id",
    #         "hh_id",
    #         "head_age",
    #         "age",
    #         "spouse_age",
    #         "ID1968",
    #         "year",
    #         "interview_number",
    #         "head_marital_status",
    #         "marital_status",
    #     ]
    # ].to_csv("psid_to_check.csv")
    # The next several lines try to identify and then drop from the sample
    # hh_ids that report more than one type of marital status
    # there are 179 of these, 26 are men who report being married and not at
    # different times, even when a spouse id is not present
    marriedmale_df = psid_df.groupby("hh_id").agg(["max"])["marriedmalehead"]
    singlemale_df = psid_df.groupby("hh_id").agg(["max"])["singlemale"]
    marriedfemale_df = psid_df.groupby("hh_id").agg(["max"])[
        "marriedfemalehead"
    ]
    singlefemale_df = psid_df.groupby("hh_id").agg(["max"])["singlefemale"]
    marriedmale_df.rename(columns={"max": "m_marriedmalehead"}, inplace=True)
    singlemale_df.rename(columns={"max": "m_singlemale"}, inplace=True)
    marriedfemale_df.rename(
        columns={"max": "m_marriedfemalehead"}, inplace=True
    )
    singlefemale_df.rename(columns={"max": "m_singlefemale"}, inplace=True)
    merged_df = marriedmale_df.join(
        [singlemale_df, marriedfemale_df, singlefemale_df],
        how="outer",
        sort=True,
    )
    merged_df["sum_status"] = (
        merged_df["m_singlemale"].astype(int)
        + merged_df["m_singlefemale"].astype(int)
        + merged_df["m_marriedfemalehead"].astype(int)
        + merged_df["m_marriedmalehead"].astype(int)
    )
    merged_df_to_list = merged_df[merged_df["sum_status"] > 1]
    merged_df_to_list.to_csv("hh_id_two_statuses.csv")
    hhid_to_drop = merged_df_to_list.copy()
    hhid_to_drop["keep"] = False
    psid_df = psid_df.merge(hhid_to_drop, on="hh_id", how="left")
    psid_df["keep"].fillna(True, inplace=True)
    psid_df = psid_df[psid_df["keep"]].copy()
    psid_df["in_psid"] = True
    # print number of obs by year
    print(
        "Number of obs by year = ",
        psid_df["hh_id"].groupby([psid_df.year]).agg("count"),
    )
    num_obs_psid = psid_df.shape[0]
    psid_df.sort_values(by=["hh_id", "year"], inplace=True)

    # clean up
    del (
        merged_df_to_list,
        hhid_to_drop,
        marriedmale_df,
        marriedfemale_df,
        singlemale_df,
        singlefemale_df,
    )

    # "fill in" observations - so have observation for each household
    # from age 20-80
    # note that do this before running regression, but that's ok since
    # wages missing here so these obs don't affect regression
    uid = psid_df["hh_id"].unique()
    all_ages = list(range(20, 81))  # for list of ages 20 to 80
    ids_full = np.array([[x] * len(all_ages) for x in list(uid)]).flatten()
    ages = all_ages * len(uid)
    balanced_panel = pd.DataFrame({"hh_id": ids_full, "age": ages})
    rebalanced_data = balanced_panel.merge(
        psid_df, how="left", on=["hh_id", "age"]
    )
    # Backfill and then forward fill variables that are constant over time
    # within hhid
    for item in PSID_CONSTANT_VARS:
        rebalanced_data[item] = rebalanced_data.groupby("hh_id")[item].fillna(
            method="bfill"
        )
        rebalanced_data[item] = rebalanced_data.groupby("hh_id")[item].fillna(
            method="ffill"
        )

    ### NOTE: we seem to get some cases where the marital status is not constant
    # despite trying to set up the indentifcation of a household such that it
    # has to be.  Why this is happening needs to be checked.

    # Fill in year by doing a cumulative counter within each hh_id and then
    # using the difference between age and this counter to infer what the
    # year should be'
    rebalanced_data.sort_values(["hh_id", "age"], inplace=True)
    rebalanced_data["counter"] = rebalanced_data.groupby("hh_id").cumcount()
    rebalanced_data["diff"] = (
        rebalanced_data["year"] - rebalanced_data["counter"]
    )
    rebalanced_data["diff"].fillna(
        0, inplace=True
    )  # because NaNs if year missing
    max_df = rebalanced_data.groupby("hh_id").agg(["max"])["diff"]
    rebalanced_data = rebalanced_data.join(max_df, how="left", on=["hh_id"])
    rebalanced_data["year"] = (
        rebalanced_data["max"] + rebalanced_data["counter"]
    )

    # clean up
    del max_df, balanced_panel

    ### Check that there are 61 obs for each hh_id

    # create additional variables for first stage regressions
    df = rebalanced_data.reset_index()
    df["age2"] = df["age"] ** 2
    df["age3"] = df["age"] ** 3
    df["age_smale"] = df["age"] * df["singlemale"]
    df["age_sfemale"] = df["age"] * df["singlefemale"]
    df["age_mmale"] = df["age"] * df["marriedmalehead"]
    df["age_mfemale"] = df["age"] * df["marriedfemalehead"]
    df["age_smale2"] = df["age2"] * df["singlemale"]
    df["age_sfemale2"] = df["age2"] * df["singlefemale"]
    df["age_mmale2"] = df["age2"] * df["marriedmalehead"]
    df["age_mfemale2"] = df["age2"] * df["marriedfemalehead"]
    df["age_smale3"] = df["age3"] * df["singlemale"]
    df["age_sfemale3"] = df["age3"] * df["singlefemale"]
    df["age_mmale3"] = df["age3"] * df["marriedmalehead"]
    df["age_mfemale3"] = df["age3"] * df["marriedfemalehead"]

    # clean up
    del rebalanced_data

    # run regressions to impute wages for years not observed in sample
    df.set_index(["hh_id", "year"], inplace=True)
    list_of_statuses = [
        "Single Males",
        "Single Females",
        "Married, Male Head",
        "Married, Female Head",
    ]
    list_of_dfs = [
        df[df["singlemale"]].copy(),
        df[df["singlefemale"]].copy(),
        df[df["marriedmalehead"]].copy(),
        df[df["marriedfemalehead"]].copy(),
    ]
    list_of_dfs_with_fitted_vals = []
    first_stage_model_results = {
        "Names": [
            "Head Age",
            "",
            "Head Age^2",
            "",
            "Head Age^3",
            "",
            "R-Squared",
            "Observations",
            "Households",
        ],
        "Single Males": [],
        "Single Females": [],
        "Married, Male Head": [],
        "Married, Female Head": [],
    }
    for i, data in enumerate(list_of_dfs):
        # Note that including entity and time effects leads to a collinearity
        # I think this is because there are some years at begin and end of
        # sample with just one person
        # mod = PanelOLS(data.ln_wage_rate,
        #                data[['age', 'age2', 'age3']],
        #                weights=data.fam_smpl_wgt_core,
        #                entity_effects=True, time_effects=True)
        mod = PanelOLS(
            data.ln_wage_rate,
            data[["age", "age2", "age3"]],
            entity_effects=True,
        )
        res = mod.fit(cov_type="clustered", cluster_entity=True)
        # print("Summary for ", list_of_statuses[i])
        # print(res.summary)
        # Save model results to dictionary
        first_stage_model_results[list_of_statuses[i]] = [
            res.params["age"],
            res.std_errors["age"],
            res.params["age2"],
            res.std_errors["age2"],
            res.params["age3"],
            res.std_errors["age3"],
            res.rsquared,
            res.nobs,
            res.entity_info["total"],
        ]
        fit_values = res.predict(fitted=True, effects=True, missing=True)
        fit_values["predictions"] = (
            fit_values["fitted_values"] + fit_values["estimated_effects"]
        )
        list_of_dfs_with_fitted_vals.append(
            data.join(fit_values, how="left", on=["hh_id", "year"])
        )

    df_w_fit = pd.concat(list_of_dfs_with_fitted_vals)
    # list_of_dfs_with_fitted_vals[0].append(
    #     list_of_dfs_with_fitted_vals[1].append(
    #         list_of_dfs_with_fitted_vals[2].append(
    #             list_of_dfs_with_fitted_vals[3]
    #         )
    #     )
    # )
    df_w_fit.rename(columns={"predictions": "ln_fillin_wage"}, inplace=True)
    # print(
    #     "Descritpion of data coming out of estimation: ", df_w_fit.describe()
    # )
    # Seems to be the same as going into estimation

    # Compute lifetime income for each filer
    int_rate = 0.04  # assumed interest rate to compute NPV of lifetime income
    time_endow = 4000
    # assumed time endowment - set at 4000 hours !!! May want
    # to change this to be different for single households than married !!!
    df_w_fit["time_wage"] = np.exp(df_w_fit["ln_fillin_wage"]) * time_endow
    df_w_fit["lifetime_inc"] = df_w_fit["time_wage"] * (
        (1 / (1 + int_rate)) ** (df_w_fit["age"] - 20)
    )
    li_df = (df_w_fit[["lifetime_inc"]].groupby(["hh_id"]).sum()).copy()
    # find percentile in distrubtion of lifetime income
    li_df["li_percentile"] = li_df.lifetime_inc.rank(pct=True)
    # Put in bins
    groups = [0.0, 0.25, 0.5, 0.7, 0.8, 0.9, 0.99, 1.0]
    cats_pct = ["0-25", "26-50", "51-70", "71-80", "81-90", "91-99", "100"]
    li_df = li_df.join(
        pd.get_dummies(pd.cut(li_df["li_percentile"], groups, labels=cats_pct))
    ).copy()
    li_df["li_group"] = pd.cut(li_df["li_percentile"], groups)
    deciles = list(np.arange(0.0, 1.1, 0.10))
    cats_10 = ["D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9", "D10"]
    li_df = li_df.join(
        pd.get_dummies(pd.cut(li_df["li_percentile"], deciles, labels=cats_10))
    ).copy()
    li_df["li_decile"] = pd.cut(li_df["li_percentile"], deciles)

    # Merge lifetime income to panel
    df_w_fit.drop(columns="lifetime_inc", inplace=True)
    df_fit2 = df_w_fit.join(
        li_df, how="left", on=["hh_id"], lsuffix="_x", rsuffix="_y"
    )
    # Drop from balanced panel those that were not in original panel
    df_fit2["in_psid"].fillna(False, inplace=True)
    panel_li = (df_fit2[df_fit2["in_psid"]]).copy()

    # Save dictionary of regression results
    pickle.dump(
        first_stage_model_results,
        open(
            os.path.join(
                CURDIR, "..", "data", "PSID", "first_stage_reg_results.pkl"
            ),
            "wb",
        ),
    )

    # Save dataframe
    # pickle.dump(panel_li, open("psid_lifetime_income.pkl", "wb"))
    panel_li.loc["li_group"] = panel_li["li_group"].astype("category")
    panel_li.loc["li_decile"] = panel_li["li_decile"].astype("category")
    panel_li.dropna(axis=0, how="all", inplace=True)
    print(panel_li.keys())
    panel_li.to_csv(
        os.path.join(CURDIR, "..", "data", "PSID", "psid_lifetime_income.csv")
    )

    return panel_li



"""
-----------------------------------------------------------------
Functions for created the matrix of ability levels, e.  This can
only be used for looking at the 25, 50, 70, 80, 90, 99, and 100th
percentiles, as it uses fitted polynomials to those percentiles.
-----------------------------------------------------------------
"""
import numpy as np
import scipy.optimize as opt
import scipy.interpolate as si
from ogcore import parameter_plots as pp
import os

CUR_PATH = os.path.abspath(os.path.dirname(__file__))
OUTPUT_DIR = os.path.join(CUR_PATH, "OUTPUT", "ability")


def arctan_func(xvals, a, b, c):
    r"""
    This function generates predicted ability levels given data (xvals)
    and parameters a, b, and c, from the following arctan function:

    .. math::
        y = (-a / \pi) * \arctan(b * x + c) + (a / 2)

    Args:
        xvals (Numpy array): data inputs to arctan function
        a (scalar): scale parameter for arctan function
        b (scalar): curvature parameter for arctan function
        c (scalar): shift parameter for arctan function

    Returns:
        yvals (Numpy array): predicted values (output) of arctan
            function

    """
    yvals = (-a / np.pi) * np.arctan(b * xvals + c) + (a / 2)
    return yvals


def arctan_deriv_func(xvals, a, b, c):
    r"""
    This function generates predicted derivatives of arctan function
    given data (xvals) and parameters a, b, and c. The functional form
    of the derivative of the function is the following:

    .. math::
        y = - (a * b) / (\pi * (1 + (b * xvals + c)^2))

    Args:
        xvals (Numpy array): data inputs to arctan derivative function
        a (scalar): scale parameter for arctan function
        b (scalar): curvature parameter for arctan function
        c (scalar): shift parameter for arctan function

    Returns:
        yvals (Numpy array): predicted values (output) of arctan
            derivative function

    """
    yvals = -(a * b) / (np.pi * (1 + (b * xvals + c) ** 2))
    return yvals


def arc_error(abc_vals, params):
    """
    This function returns a vector of errors in the three criteria on
    which the arctan function is fit to predict extrapolated ability in
    ages 81 to 100.::

        1) The arctan function value at age 80 must match the estimated
           original function value at age 80.
        2) The arctan function slope at age 80 must match the estimated
           original function slope at age 80.
        3) The level of ability at age 100 must be a given fraction
           (abil_deprec) below the ability level at age 80.

    Args:
        abc_vals (tuple): contains (a,b,c)

            * a (scalar): scale parameter for arctan function
            * b (scalar): curvature parameter for arctan function
            * c (scalar): shift parameter for arctan function
        params (tuple): contains (first_point, coef1, coef2, coef3,
            abil_deprec)

            * first_point (scalar): ability level at age 80, > 0
            * coef1 (scalar): coefficient in log ability equation on
                linear term in age
            * coef2 (scalar): coefficient in log ability equation on
                quadratic term in age
            * coef3 (scalar): coefficient in log ability equation on
                cubic term in age
            * abil_deprec (scalar): ability depreciation rate between
                ages 80 and 100, in (0, 1).

    Returns:
        error_vec (Numpy array): errors ([error1, error2, error3])

            * error1 (scalar): error between ability level at age 80
                from original function minus the predicted ability at
                age 80 from the arctan function given a, b, and c
            * error2 (scalar): error between the slope of the original
                function at age 80 minus the slope of the arctan
                function at age 80 given a, b, and c
            * error3 (scalar): error between the ability level at age
                100 predicted by the original model value times
                abil_deprec minus the ability predicted by the arctan
                function at age 100 given a, b, and c

    """
    a, b, c = abc_vals
    first_point, coef1, coef2, coef3, abil_deprec = params
    error1 = first_point - arctan_func(80, a, b, c)
    if (3 * coef3 * 80**2 + 2 * coef2 * 80 + coef1) < 0:
        error2 = (
            3 * coef3 * 80**2 + 2 * coef2 * 80 + coef1
        ) * first_point - arctan_deriv_func(80, a, b, c)
    else:
        error2 = -0.02 * first_point - arctan_deriv_func(80, a, b, c)
    error3 = abil_deprec * first_point - arctan_func(100, a, b, c)
    error_vec = np.array([error1, error2, error3])

    return error_vec


def arctan_fit(first_point, coef1, coef2, coef3, abil_deprec, init_guesses):
    """
    This function fits an arctan function to the last 20 years of the
    ability levels of a particular ability group to extrapolate
    abilities by trying to match the slope in the 80th year and the
    ability depreciation rate between years 80 and 100.

    Args:
        first_point (scalar): ability level at age 80, > 0
        coef1 (scalar): coefficient in log ability equation on linear
            term in age
        coef2 (scalar): coefficient in log ability equation on
            quadratic term in age
        coef3 (scalar): coefficient in log ability equation on cubic
            term in age
        abil_deprec (scalar): ability depreciation rate between
            ages 80 and 100, in (0, 1)
        init_guesses (Numpy array): initial guesses

    Returns:
        abil_last (Numpy array): extrapolated ability levels for ages
            81 to 100, length 20

    """
    params = [first_point, coef1, coef2, coef3, abil_deprec]
    solution = opt.root(arc_error, init_guesses, args=params, method="lm")
    [a, b, c] = solution.x
    old_ages = np.linspace(81, 100, 20)
    abil_last = arctan_func(old_ages, a, b, c)
    return abil_last


def get_e_interp(S, age_wgts, age_wgts_80, abil_wgts, plot=False):
    """
    This function takes a source matrix of lifetime earnings profiles
    (abilities, emat) of size (80, 7), where 80 is the number of ages
    and 7 is the number of ability types in the source matrix, and
    interpolates new values of a new S x J sized matrix of abilities
    using linear interpolation. [NOTE: For this application, cubic
    spline interpolation introduces too much curvature.]

    This function also includes the two cases in which J = 9 and J = 10
    that include higher lifetime earning percentiles calibrated using
    Piketty and Saez (2003).


    Args:
        S (int): number of ages to interpolate. This method assumes that
            ages are evenly spaced between the beginning of the 21st
            year and the end of the 100th year, >= 3
        age_wgts (Numpy array): distribution of population in each age
            for the interpolated ages, length S
        age_wgts_80 (Numpy array): percent of population in each
            one-year age from 21 to 100, length 80
        abil_wgts (Numpy array): distribution of population in each
            ability group, length J
        plot (bool): if True, creates plots of emat_orig and the new
            interpolated emat_new

    Returns:
        emat_new_scaled (Numpy array): interpolated ability matrix scaled
            so that population-weighted average is 1, size SxJ

    """
    # Get original 80 x 7 ability matrix
    abil_wgts_orig = np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.01])
    emat_orig = get_e_orig(age_wgts_80, abil_wgts_orig, plot)
    if (
        S == 80
        and np.array_equal(
            np.squeeze(abil_wgts),
            np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.01]),
        )
        is True
    ):
        emat_new_scaled = emat_orig
    elif (
        S == 80
        and np.array_equal(
            np.squeeze(abil_wgts),
            np.array(
                [0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.005, 0.004, 0.0009, 0.0001]
            ),
        )
        is True
    ):
        emat_new = np.zeros((S, len(abil_wgts)))
        emat_new[:, :7] = emat_orig
        # Create profiles for top 0.5%, top 0.1% and top 0.01% using
        # Piketty and Saez estimates
        # (https://eml.berkeley.edu/~saez/pikettyqje.pdf)
        # updated for 2018 to create scaling factor
        # assumption is that profile shape of these top 3 groups are
        # same as the top 1% estimated in tax data, just scaled up by
        # ratio determined from P&S 2018 estimates (Table 0, ex cap gains)
        emat_new[:, 5] = emat_orig[:, -2] * 1.25
        emat_new[:, 6] = emat_orig[:, -1] * 0.458759521 * 2.75
        emat_new[:, 7] = emat_orig[:, -1] * 0.847252448 * 3.5
        emat_new[:, 8] = emat_orig[:, -1] * 2.713698465 * 3.5
        emat_new[:, 9] = emat_orig[:, -1] * 18.74863983 * 4.0
        emat_new_scaled = (
            emat_new
            / (
                emat_new * age_wgts.reshape(80, 1) * abil_wgts.reshape(1, 10)
            ).sum()
        )
    elif (
        S == 80
        and np.array_equal(
            np.squeeze(abil_wgts),
            np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.005, 0.004, 0.001]),
        )
        is True
    ):
        emat_new = np.zeros((S, len(abil_wgts)))
        emat_new[:, :7] = emat_orig
        # Create profiles for top 0.5%, top 0.1% using
        # Piketty and Saez estimates
        # (https://eml.berkeley.edu/~saez/pikettyqje.pdf)
        # updated for 2018 to create scaling factor
        # assumption is that profile shape of these top 3 groups are
        # same as the top 1% estimated in tax data, just scaled up by
        # ratio determined from P&S 2018 estimates (Table 0, ex cap gains)
        emat_new[:, 6] = emat_orig[:, -1] * 0.458759521
        emat_new[:, 7] = emat_orig[:, -1] * 0.847252448
        emat_new[:, 8] = emat_orig[:, -1] * 4.317192601
        emat_new_scaled = (
            emat_new
            / (
                emat_new * age_wgts.reshape(80, 1) * abil_wgts.reshape(1, 9)
            ).sum()
        )
    else:
        # generate abil_midp vector
        J = abil_wgts.shape[0]
        abil_midp = np.zeros(J)
        pct_lb = 0.0
        for j in range(J):
            abil_midp[j] = pct_lb + 0.5 * abil_wgts[j]
            pct_lb += abil_wgts[j]

        # Make sure that values in abil_midp are within interpolating
        # bounds set by the hard coded abil_wgts_orig
        if abil_midp.min() < 0.125 or abil_midp.max() > 0.995:
            err = (
                "One or more entries in abils vector is outside the "
                + "allowable bounds."
            )
            raise RuntimeError(err)

        emat_j_midp = np.array(
            [0.125, 0.375, 0.600, 0.750, 0.850, 0.945, 0.995]
        )
        emat_s_midp = np.linspace(20.5, 99.5, 80)
        emat_j_mesh, emat_s_mesh = np.meshgrid(emat_j_midp, emat_s_midp)
        newstep = 80 / S
        new_s_midp = np.linspace(20 + 0.5 * newstep, 100 - 0.5 * newstep, S)
        new_j_mesh, new_s_mesh = np.meshgrid(abil_midp, new_s_midp)
        newcoords = np.hstack(
            (
                emat_s_mesh.reshape((80 * 7, 1)),
                emat_j_mesh.reshape((80 * 7, 1)),
            )
        )
        emat_new = si.griddata(
            newcoords,
            emat_orig.flatten(),
            (new_s_mesh, new_j_mesh),
            method="linear",
        )
        emat_new_scaled = (
            emat_new
            / (
                emat_new * age_wgts.reshape(S, 1) * abil_wgts.reshape(1, J)
            ).sum()
        )

        if plot:
            kwargs = {"filesuffix": "_intrp_scaled"}
            pp.plot_income_data(
                new_s_midp,
                abil_midp,
                abil_wgts,
                emat_new_scaled,
                OUTPUT_DIR,
                **kwargs,
            )

    return emat_new_scaled


def get_e_orig(age_wgts, abil_wgts, plot=False):
    r"""
    This function generates the 80 x 7 matrix of lifetime earnings
    ability profiles, corresponding to annual ages from 21 to 100 and to
    paths based on income percentiles 0-25, 25-50, 50-70, 70-80, 80-90,
    90-99, 99-100. The ergodic population distribution is an input in
    order to rescale the paths so that the weighted average equals 1.

    The data come from the following file:

        `data/ability/FR_wage_profile_tables.xlsx`

    The polynomials are of the form

    .. math::
        \ln(abil) = \alpha + \beta_{1}\text{age} + \beta_{2}\text{age}^2
            + \beta_{3}\text{age}^3

    Values come from regression analysis using IRS CWHS with hours
    imputed from the CPS.

    Args:
        age_wgts (Numpy array): ergodic age distribution, length S
        abil_wgts (Numpy array): population weights in each lifetime
            earnings group, length J
        plot (bool): if True, generates 3D plots of ability paths

    Returns:
        e_orig_scaled (Numpy array): = lifetime ability profiles scaled
            so that population-weighted average is 1, size SxJ

    """
    # Return and error if age_wgts is not a vector of size (80,)
    if age_wgts.shape[0] != 80:
        err = "Vector age_wgts does not have 80 elements."
        raise RuntimeError(err)
    # Return and error if abil_wgts is not a vector of size (7,)
    if abil_wgts.shape[0] != 7:
        err = "Vector abil_wgts does not have 7 elements."
        raise RuntimeError(err)

    # 1) Generate polynomials and use them to get income profiles for
    #    ages 21 to 80.
    one = np.array(
        [
            -0.09720122,
            0.05995294,
            0.17654618,
            0.21168263,
            0.21638731,
            0.04500235,
            0.09229392,
        ]
    )
    two = np.array(
        [
            0.00247639,
            -0.00004086,
            -0.00240656,
            -0.00306555,
            -0.00321041,
            0.00094253,
            0.00012902,
        ]
    )
    three = np.array(
        [
            -0.00001842,
            -0.00000521,
            0.00001039,
            0.00001438,
            0.00001579,
            -0.00001470,
            -0.00001169,
        ]
    )
    const = np.array(
        [
            3.41e00,
            0.69689692,
            -0.78761958,
            -1.11e00,
            -0.93939272,
            1.60e00,
            1.89e00,
        ]
    )
    ages_short = np.tile(np.linspace(21, 80, 60).reshape((60, 1)), (1, 7))
    log_abil_paths = (
        const
        + (one * ages_short)
        + (two * (ages_short**2))
        + (three * (ages_short**3))
    )
    abil_paths = np.exp(log_abil_paths)
    e_orig = np.zeros((80, 7))
    e_orig[:60, :] = abil_paths
    e_orig[60:, :] = 0.0

    # 2) Forecast (with some art) the path of the final 20 years of
    #    ability types. This following variable is what percentage of
    #    ability at age 80 ability falls to at age 100. In general, we
    #    wanted people to lose half of their ability over a 20-year
    #    period. The first entry is 0.47, though, because nothing higher
    #    would converge. The second-to-last is 0.7 because this group
    #    actually has a slightly higher ability at age 80 than the last
    #    group, so this value makes it decrease more so it ends up being
    #    monotonic.
    abil_deprec = np.array([0.47, 0.5, 0.5, 0.5, 0.5, 0.7, 0.5])
    #     Initial guesses for the arctan. They're pretty sensitive.
    init_guesses = np.array(
        [
            [58, 0.0756438545595, -5.6940142786],
            [27, 0.069, -5],
            [35, 0.06, -5],
            [37, 0.339936555352, -33.5987329144],
            [70.5229181668, 0.0701993896947, -6.37746859905],
            [35, 0.06, -5],
            [35, 0.06, -5],
        ]
    )
    for j in range(7):
        e_orig[60:, j] = arctan_fit(
            e_orig[59, j],
            one[j],
            two[j],
            three[j],
            abil_deprec[j],
            init_guesses[j],
        )

    # 3) Rescale the lifetime earnings path matrix so that the
    #    population weighted average equals 1.
    e_orig_scaled = (
        e_orig
        / (e_orig * age_wgts.reshape(80, 1) * abil_wgts.reshape(1, 7)).sum()
    )

    if plot:
        ages_long = np.linspace(21, 100, 80)
        abil_midp = np.array([12.5, 37.5, 60.0, 75.0, 85.0, 94.5, 99.5])
        # Plot original unscaled 80 x 7 ability matrix
        kwargs = {"filesuffix": "_orig_unscaled"}
        pp.plot_income_data(
            ages_long, abil_midp, abil_wgts, e_orig, OUTPUT_DIR, **kwargs
        )

        # Plot original scaled 80 x 7 ability matrix
        kwargs = {"filesuffix": "_orig_scaled"}
        pp.plot_income_data(
            ages_long,
            abil_midp,
            abil_wgts,
            e_orig_scaled,
            OUTPUT_DIR,
            **kwargs,
        )

    return e_orig_scaled



import pandas as pd
import numpy as np
import pickle
import os


# Create directory if output directory does not already exist
cur_path = os.path.split(os.path.abspath(__file__))[0]
output_fldr = "csv_output_files"
output_dir = os.path.join(cur_path, "..", "data", "PSID", output_fldr)
if not os.access(output_dir, os.F_OK):
    os.makedirs(output_dir)

# Read in dictionary of regression results
first_stage_results = pickle.load(
    open(os.path.join(cur_path, "first_stage_reg_results.pkl"), "rb")
)

# Read in dataframe of PSID data
df = pickle.load(
    open(
        os.path.join(
            cur_path, "..", "data", "PSID", "psid_lifetime_income.pkl"
        ),
        "rb",
    )
)

# save psid to stata data file
stata_df = df.copy()
stata_df["li_group"] = stata_df.li_group.astype(str)
stata_df["li_decile"] = stata_df.li_decile.astype(str)
stata_df.to_csv("psid_lifetime_income.csv", index=False)

# Put regression results in a table to be read into excel for formatting
first_stage_reg = pd.DataFrame.from_dict(first_stage_results)
first_stage_reg.to_csv(
    os.path.join(output_dir, "FirstStageRegressionTable.csv"), header=True
)

# Compute summary stats and save to tables


# /* Summary Statistics for full sample */
df["wage_21"] = np.nan
df["wage_50"] = np.nan
df["wage_80"] = np.nan
df["earn_21"] = np.nan
df["earn_50"] = np.nan
df["earn_80"] = np.nan
df.loc[df.age == 21, "wage_21"] = df["wage_rate"]
df.loc[df.age == 50, "wage_50"] = df["wage_rate"]
df.loc[df.age == 80, "wage_80"] = df["wage_rate"]
df.loc[df.age == 21, "earn_21"] = df["earn_rate"]
df.loc[df.age == 50, "earn_50"] = df["earn_rate"]
df.loc[df.age == 80, "earn_80"] = df["earn_rate"]

vars_to_collapse = [
    "wage_21",
    "wage_50",
    "wage_80",
    "earn_21",
    "earn_50",
    "earn_80",
    "wage_rate",
    "earn_rate",
    "businc_hh",
    "lifetime_inc",
    "earninc_hh",
    "earnhours_hh",
    "incwage_hh",
    "married",
    "avg_earn_inc",
    "avg_wage_inc",
    "head_age",
    "singlefemale",
    "singlemale",
    "marriedmalehead",
    "marriedfemalehead",
]
df.singlefemale = df.singlefemale.astype(int)
df.singlemale = df.singlemale.astype(int)
df.marriedfemalehead = df.marriedfemalehead.astype(int)
df.marriedmalehead = df.marriedmalehead.astype(int)

mean_df = df[["earninc_hh", "incwage_hh"]].groupby("hh_id").mean()
mean_df.rename(
    columns={"earninc_hh": "avg_earn_inc", "incwage_hh": "avg_wage_inc"},
    inplace=True,
)
df = df.merge(mean_df, how="left", on="hh_id", copy=True)
fullsample_means = df[vars_to_collapse].mean()
fullsample_sds = df[vars_to_collapse].std()
fullsample_counts = df[vars_to_collapse].count()
fullsample_means.to_csv(
    os.path.join(output_dir, "full_sample_means.csv"), header=False
)
fullsample_sds.to_csv(
    os.path.join(output_dir, "full_sample_sds.csv"), header=False
)
fullsample_counts.to_csv(
    os.path.join(output_dir, "full_sample_counts.csv"), header=False
)

# Summary Stats for full sample - but with one obs per filer
one_per = df.groupby("hh_id").mean()
fullsample1_means = one_per[vars_to_collapse].mean()
fullsample1_sds = one_per[vars_to_collapse].std()
fullsample1_counts = one_per[vars_to_collapse].count()
fullsample1_means.to_csv(
    os.path.join(output_dir, "full_sample_means_one_per_filer.csv"),
    header=False,
)
fullsample1_sds.to_csv(
    os.path.join(output_dir, "full_sample_sds_one_per_filer.csv"), header=False
)
fullsample1_counts.to_csv(
    os.path.join(output_dir, "full_sample_counts_one_per_filer.csv"),
    header=False,
)

# Summary Stats for full sample by lifetime income decile
decile_means = df[vars_to_collapse + ["li_decile"]].groupby("li_decile").mean()
decile_sds = df[vars_to_collapse + ["li_decile"]].groupby("li_decile").std()
decile_counts = (
    df[vars_to_collapse + ["li_decile"]].groupby("li_decile").count()
)
decile_means.to_csv(
    os.path.join(output_dir, "means_by_li_decile.csv"), header=False
)
decile_sds.to_csv(
    os.path.join(output_dir, "sds_by_li_decile.csv"), header=False
)
decile_counts.to_csv(
    os.path.join(output_dir, "counts_by_li_decile.csv"), header=False
)

# Table of descriptive statistics by lifetime income percentile group
group_means = df[vars_to_collapse + ["li_group"]].groupby("li_group").mean()
group_sds = df[vars_to_collapse + ["li_group"]].groupby("li_group").std()
group_counts = df[vars_to_collapse + ["li_group"]].groupby("li_group").count()
group_means.to_csv(
    os.path.join(output_dir, "means_by_li_group.csv"), header=False
)
group_sds.to_csv(os.path.join(output_dir, "sds_by_li_group.csv"), header=False)
group_counts.to_csv(
    os.path.join(output_dir, "counts_by_li_group.csv"), header=False
)

# Put together formatted descriptive stats tables
cats_pct = ["0-25", "26-50", "51-70", "71-80", "81-90", "91-99", "100"]
summ_stats_by_group = {
    "Lifetime Income Percentile:": cats_pct + ["All"],
    "Observations": [],
    "Fraction Single": [""] * (len(cats_pct) + 1),
    "   Females": [],
    "   Males": [],
    "Fraction Married": [""] * (len(cats_pct) + 1),
    "   Female Head": [],
    "   Male Head": [],
    "Means:": [""] * (len(cats_pct) + 1),
    "   Age, Head": [],
    "   Hourly Wage": [],
    "   Annual Wages": [],
    "   Lifetime Income": [],
    "   Hours": [],
    "   Hourly Wage, Age 21": [],
    "   Hourly Wage, Age 50": [],
    "   Hourly Wage, Age 80": [],
}
# Append values for each group
for i, v in enumerate(cats_pct):
    summ_stats_by_group["Observations"].append(
        group_counts.loc[i, "lifetime_inc"]
    )
    summ_stats_by_group["   Females"].append(
        group_means.loc[i, "singlefemale"]
    )
    summ_stats_by_group["   Males"].append(group_means.loc[i, "singlemale"])
    summ_stats_by_group["   Female Head"].append(
        group_means.loc[i, "marriedfemalehead"]
    )
    summ_stats_by_group["   Male Head"].append(
        group_means.loc[i, "marriedmalehead"]
    )
    summ_stats_by_group["   Age, Head"].append(group_means.loc[i, "head_age"])
    summ_stats_by_group["   Hourly Wage"].append(
        group_means.loc[i, "earn_rate"]
    )
    summ_stats_by_group["   Annual Wages"].append(
        group_means.loc[i, "earninc_hh"]
    )
    summ_stats_by_group["   Lifetime Income"].append(
        group_means.loc[i, "lifetime_inc"]
    )
    summ_stats_by_group["   Hours"].append(group_means.loc[i, "earnhours_hh"])
    summ_stats_by_group["   Hourly Wage, Age 21"].append(
        group_means.loc[i, "earn_21"]
    )
    summ_stats_by_group["   Hourly Wage, Age 50"].append(
        group_means.loc[i, "earn_50"]
    )
    summ_stats_by_group["   Hourly Wage, Age 80"].append(
        group_means.loc[i, "earn_80"]
    )
# Append averages
summ_stats_by_group["Observations"].append(fullsample_counts["lifetime_inc"])
summ_stats_by_group["   Females"].append(fullsample_means["singlefemale"])
summ_stats_by_group["   Males"].append(fullsample_means["singlemale"])
summ_stats_by_group["   Female Head"].append(
    fullsample_means["marriedfemalehead"]
)
summ_stats_by_group["   Male Head"].append(fullsample_means["marriedmalehead"])
summ_stats_by_group["   Age, Head"].append(fullsample_means["head_age"])
summ_stats_by_group["   Hourly Wage"].append(fullsample_means["earn_rate"])
summ_stats_by_group["   Annual Wages"].append(fullsample_means["earninc_hh"])
summ_stats_by_group["   Lifetime Income"].append(
    fullsample_means["lifetime_inc"]
)
summ_stats_by_group["   Hours"].append(fullsample_means["earnhours_hh"])
summ_stats_by_group["   Hourly Wage, Age 21"].append(
    fullsample_means["earn_21"]
)
summ_stats_by_group["   Hourly Wage, Age 50"].append(
    fullsample_means["earn_50"]
)
summ_stats_by_group["   Hourly Wage, Age 80"].append(
    fullsample_means["earn_80"]
)
summ_group = pd.DataFrame.from_dict(summ_stats_by_group)
print(summ_group)
summ_group.to_csv(os.path.join(output_dir, "summ_stats_by_group.csv"))

decile_list = list(range(1, 11))
summ_stats_by_decile = {
    "Lifetime Income Decile:": decile_list + ["All"],
    "Observations": [],
    "Fraction Single": [""] * (len(decile_list) + 1),
    "   Females": [],
    "   Males": [],
    "Fraction Married": [""] * (len(decile_list) + 1),
    "   Female Head": [],
    "   Male Head": [],
    "Means:": [""] * (len(decile_list) + 1),
    "   Age, Head": [],
    "   Hourly Wage": [],
    "   Annual Wages": [],
    "   Lifetime Income": [],
    "   Hours": [],
    "   Hourly Wage, Age 21": [],
    "   Hourly Wage, Age 50": [],
    "   Hourly Wage, Age 80": [],
}
# Append values for each decile
for i, v in enumerate(decile_list):
    summ_stats_by_decile["Observations"].append(
        decile_counts.loc[i, "lifetime_inc"]
    )
    summ_stats_by_decile["   Females"].append(
        decile_means.loc[i, "singlefemale"]
    )
    summ_stats_by_decile["   Males"].append(decile_means.loc[i, "singlemale"])
    summ_stats_by_decile["   Female Head"].append(
        decile_means.loc[i, "marriedfemalehead"]
    )
    summ_stats_by_decile["   Male Head"].append(
        decile_means.loc[i, "marriedmalehead"]
    )
    summ_stats_by_decile["   Age, Head"].append(
        decile_means.loc[i, "head_age"]
    )
    summ_stats_by_decile["   Hourly Wage"].append(
        decile_means.loc[i, "earn_rate"]
    )
    summ_stats_by_decile["   Annual Wages"].append(
        decile_means.loc[i, "earninc_hh"]
    )
    summ_stats_by_decile["   Lifetime Income"].append(
        decile_means.loc[i, "lifetime_inc"]
    )
    summ_stats_by_decile["   Hours"].append(
        decile_means.loc[i, "earnhours_hh"]
    )
    summ_stats_by_decile["   Hourly Wage, Age 21"].append(
        decile_means.loc[i, "earn_21"]
    )
    summ_stats_by_decile["   Hourly Wage, Age 50"].append(
        decile_means.loc[i, "earn_50"]
    )
    summ_stats_by_decile["   Hourly Wage, Age 80"].append(
        decile_means.loc[i, "earn_80"]
    )
# Append averages
summ_stats_by_decile["Observations"].append(fullsample_counts["lifetime_inc"])
summ_stats_by_decile["   Females"].append(fullsample_means["singlefemale"])
summ_stats_by_decile["   Males"].append(fullsample_means["singlemale"])
summ_stats_by_decile["   Female Head"].append(
    fullsample_means["marriedfemalehead"]
)
summ_stats_by_decile["   Male Head"].append(
    fullsample_means["marriedmalehead"]
)
summ_stats_by_decile["   Age, Head"].append(fullsample_means["head_age"])
summ_stats_by_decile["   Hourly Wage"].append(fullsample_means["earn_rate"])
summ_stats_by_decile["   Annual Wages"].append(fullsample_means["earninc_hh"])
summ_stats_by_decile["   Lifetime Income"].append(
    fullsample_means["lifetime_inc"]
)
summ_stats_by_decile["   Hours"].append(fullsample_means["earnhours_hh"])
summ_stats_by_decile["   Hourly Wage, Age 21"].append(
    fullsample_means["earn_21"]
)
summ_stats_by_decile["   Hourly Wage, Age 50"].append(
    fullsample_means["earn_50"]
)
summ_stats_by_decile["   Hourly Wage, Age 80"].append(
    fullsample_means["earn_80"]
)

summ_decile = pd.DataFrame.from_dict(summ_stats_by_decile)
print(summ_decile)
summ_decile.to_csv(os.path.join(output_dir, "summ_stats_by_decile.csv"))



"""
------------------------------------------------------------------------
Computes the average labor participation rate for each age cohort.
------------------------------------------------------------------------
"""

import os
import numpy as np
import pandas as pd
import scipy.ndimage.filters as filter
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


"""
------------------------------------------------------------------------
    Import Data
------------------------------------------------------------------------
"""


def get_labor_data():
    """
    Read in "raw" CPS data to calculate moments.  These data were
    cleaned in hours_data_cps_setup.do
    fileDir = os.path.dirname(os.path.realpath('__file__'))
    filename = os.path.join(
        fileDir,
        '../Data/Current_Population_Survey/cps_est_ability_hours_1992to2013.dta')
    filename = os.path.abspath(os.path.realpath(filename))
    cps = pd.read_stata(filename, columns=['year', 'age', 'hours',
                                           'hours_unit', 'wtsupp'])

    Args:
        None

    Returns:
        cps (Pandas DataFrame): CPS data to compute labor supply from

    """


#     Need to:
#     1) read in raw CPS files
#     2) do collapsing
#     3) return pandas DF with raw CPS data (just variables needed -
#        age, hours, weight)
#     4) return np array "weighted"
#
#     5) moments() will take CPS and calc moments
#     6) VCV will boot strap CPS and call moments() with each
#        boostrapped sample
# '''
#
#     # Create variables for number of age groups in data (S_labor) and number
#     # of percentiles (J_labor)
#     S_labor = 60
#     J_labor = 99
#
#     labor_file = utils.read_file(cur_path,
#                                  "data/labor/cps_hours_by_age_hourspct.txt")
#     data = pd.read_csv(labor_file, header=0)
#
#     piv = data.pivot(index='age', columns='hours_pct', values='mean_hrs')
#     lab_mat_basic = np.array(piv)
#     lab_mat_basic /= np.nanmax(lab_mat_basic)
#
#     piv2 = data.pivot(index='age', columns='hours_pct', values='num_obs')
#     weights = np.array(piv2)
#     weights /= np.nansum(weights, axis=1).reshape(S_labor, 1)
#     weighted = np.nansum((lab_mat_basic * weights), axis=1)


def compute_labor_moments(cps, S):
    """
    Compute moments from labor data.

    Args:
        cps (Pandas DataFrame): CPS data to compute labor supply from
        S (int): number of periods of economic life for model households

    Returns:
        labor_dist_out (Numpy array): fraction of time spent working
            by age

    """

    # Find fraction of total time people work on average by age
    cps["hours_wgt"] = cps["hours"] * cps["wtsupp"]
    columns = ["hours_wgt", "wgt", "avg_hours"]
    by_age = pd.DataFrame(columns=columns)
    # by_age = by_age.fillna(0) # with 0s rather than NaNs

    by_age["hours_wgt"] = cps.groupby(["age"])["hours_wgt"].sum()
    by_age["wgt"] = cps.groupby(["age"])["wtsupp"].sum()
    by_age["avg_hours"] = by_age["hours_wgt"] / by_age["wgt"]

    # get fraction of time endowment worked (assume time
    # endowment is 24 hours minus required time to sleep)
    by_age["frac_work"] = by_age["avg_hours"] / (365 * 16.0)

    # Data have sufficient obs through age  80
    # Fit a line to the last few years of the average labor
    # participation which extends from ages 76 to 100.
    slope = (by_age["frac_work"][-1] - by_age["frac_work"][-15]) / (15.0)
    # intercept = by_age['frac_work'][-1] - slope*len(by_age['frac_work'])
    # extension = slope * (np.linspace(56, 80, 23)) + intercept
    # to_dot = slope * (np.linspace(45, 56, 11)) + intercept

    labor_dist_data = np.zeros(80)
    labor_dist_data[:60] = by_age["frac_work"]
    labor_dist_data[60:] = by_age["frac_work"][-1] + slope * range(20)

    # the above computes moments if the model period is a year
    # the following adjusts those moments in case it is smaller
    labor_dist_out = filter.uniform_filter(labor_dist_data, size=int(80 / S))[
        :: int(80 / S)
    ]

    return labor_dist_out


def VCV_moments(cps, n, bin_weights, S):
    """
    Compute Variance-Covariance matrix for labor moments by
    bootstrapping data.

    Args:
        cps (Pandas DataFrame): CPS data to compute labor supply from
        S (int): number of periods of economic life for model households
        n (int): number of bootstrap iterations to run
        bin_weights (Numpy array): ability weight, length J

    Output:
        VCV (Numpy array): = variance-covariance matrix of labor
            moments, size SxS

    """
    labor_moments_boot = np.zeros((n, S))
    for i in range(n):
        boot = cps[np.random.randint(2, size=len(cps.index)).astype(bool)]
        labor_moments_boot[i, :] = compute_labor_moments(boot, S)

    VCV = np.cov(labor_moments_boot.T)

    return VCV


def labor_data_graphs(weighted, S, J, output_dir):
    """
    Plot labor supply data.

    Args:
        weighted (Numpy array):
        S (int): number of periods of economic life for model households
        J (int): number of lifetime income groups
        output_dir (str): path to save figures to

    Returns:
        None

    """
    domain = np.linspace(20, 80, S)
    Jgrid = np.linspace(1, 100, J)
    X, Y = np.meshgrid(domain, Jgrid)
    cmap2 = matplotlib.cm.get_cmap("summer")

    plt.plot(domain, weighted, color="black", label="Data")
    plt.plot(
        np.linspace(76, 100, 23),
        extension,
        color="black",
        linestyle="-.",
        label="Extrapolation",
    )
    plt.plot(np.linspace(65, 76, 11), to_dot, linestyle="--", color="black")
    plt.axvline(x=76, color="black", linestyle="--")
    plt.xlabel(r"age-$s$")
    plt.ylabel(r"individual labor supply $/bar{l}_s$")
    plt.legend()
    plt.savefig(
        os.path.join(baseline_dir, "Demographics/labor_dist_data_withfit.png")
    )

    fig10 = plt.figure()
    ax10 = fig10.add_subplot(projection="3d")
    ax10.plot_surface(X, Y, lab_mat_basic.T, rstride=1, cstride=2, cmap=cmap2)
    ax10.set_xlabel(r"age-$s$")
    ax10.set_ylabel(r"ability type -$j$")
    ax10.set_zlabel(r"labor $e_j(s)$")
    plt.savefig(os.path.join(baseline_dir, "Demographics/data_labor_dist"))



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
from ogusa.utils import MVKDE

CURDIR = os.path.split(os.path.abspath(__file__))[0]


def get_transfer_matrix(
    J=7, lambdas=np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.01]), graphs=True
):
    """
    Compute SxJ matrix representing the distribution of aggregate
    government transfers by age and lifetime income group.
    """
    # Create directory if output directory does not already exist
    CURDIR = os.path.split(os.path.abspath(__file__))[0]
    output_fldr = "csv_output_files"
    output_dir = os.path.join(CURDIR, output_fldr)
    if not os.access(output_dir, os.F_OK):
        os.makedirs(output_dir)
    image_fldr = "images"
    image_dir = os.path.join(CURDIR, image_fldr)
    if not os.access(image_dir, os.F_OK):
        os.makedirs(image_dir)

    # Define a lambda function to compute the weighted mean:
    # wm = lambda x: np.average(
    #     x, weights=df.loc[x.index, "fam_smpl_wgt_core"])

    # Read in dataframe of PSID data
    # df = ogcore.utils.safe_read_pickle(
    #     os.path.join(CURDIR, "data", "PSID", "psid_lifetime_income.pkl")
    # )
    df = pd.read_csv(
        os.path.join(CURDIR, "..", "data", "PSID", "psid_lifetime_income.csv")
    )

    # Do some tabs with data file...
    df["total_transfers"] = (
        df["head_and_spouse_transfer_income"]
        + df["other_familyunit_transfer_income"]
    )

    df["sum_transfers"] = (
        # df["other_familyunit_ssi_prior_year"] # don't include SSI since OG-USA models separately
        df["head_other_welfare_prior_year"]
        + df["spouse_other_welfare_prior_year"]
        + df["other_familyunit_other_welfare_prior_year"]
        + df["head_unemp_inc_prior_year"]
        + df["spouse_unemp_inc_prior_year"]
        + df["other_familyunit_unemp_inc_prior_year"]
    )

    if graphs:
        # Total total_transfers by year
        df.groupby("year_data").mean(numeric_only=True).plot(
            y="total_transfers"
        )
        plt.savefig(os.path.join(image_dir, "total_transfers_year.png"))
        df.groupby("year_data").mean(numeric_only=True).plot(y="sum_transfers")
        plt.savefig(os.path.join(image_dir, "sum_transfers_year.png"))
        # note that the sum of transfer categories is much lower than the
        # "total transfers" variable.  The transfers variable goes more to high income
        # and old, even though it says it excludes social security
        # because of this, we'll use the "sum transfers" variable

        # Fraction of total_transfers in a year by age
        # line plot
        df[df["year_data"] >= 1988].groupby("age").mean(
            numeric_only=True
        ).plot(y="total_transfers")
        plt.savefig(os.path.join(image_dir, "total_transfers_age.png"))

        # total_transfers by lifetime income group
        # bar plot
        df[df["year_data"] >= 1988].groupby("li_group").mean(
            numeric_only=True
        ).plot.bar(y="total_transfers")
        plt.savefig(os.path.join(image_dir, "total_transfers_li.png"))

        # lifecycle plots with line for each ability type
        pd.pivot_table(
            df[df["year_data"] >= 1988],
            values="total_transfers",
            index="age",
            columns="li_group",
            aggfunc="mean",
        ).plot(legend=True)
        plt.savefig(os.path.join(image_dir, "total_transfers_age_li.png"))

        pd.pivot_table(
            df[df["year_data"] >= 1988],
            values="sum_transfers",
            index="age",
            columns="li_group",
            aggfunc="mean",
        ).plot(legend=True)
        plt.savefig(os.path.join(image_dir, "sum_transfers_age_li.png"))

    # Matrix Fraction of sum_transfers in a year by age and lifetime_inc
    transfers_matrix = pd.pivot_table(
        df[df["year_data"] >= 1988],
        values="sum_transfers",
        index="age",
        columns="li_group",
        aggfunc="sum",
    )
    # replace NaN with zero
    transfers_matrix.fillna(value=0, inplace=True)
    transfers_matrix = transfers_matrix / transfers_matrix.sum().sum()
    # total_transfers_matrix.to_csv(os.path.join(
    #     output_dir, 'transfer_matrix.csv'))

    # estimate kernel density of transfers
    kde_matrix = MVKDE(
        80,
        7,
        transfers_matrix.to_numpy(),
        filename=os.path.join(image_dir, "sum_transfers_kde.png"),
        plot=True,
        bandwidth=0.5,
    )

    if (J == 10) and np.array_equal(
        np.squeeze(lambdas[:6]), np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09])
    ):
        kde_matrix_new = np.zeros((80, J))
        kde_matrix_new[:, :6] = kde_matrix[:, :6]
        kde_matrix_new[:, 6:] = (
            kde_matrix[:, 6:].sum(axis=1).reshape(80, 1)
            * np.tile(np.reshape(lambdas[6:], (1, 4)), (80, 1))
            / lambdas[6:].sum()
        )
        kde_matrix = kde_matrix_new

    np.savetxt(
        os.path.join(output_dir, "sum_transfers_kde.csv"),
        kde_matrix,
        delimiter=",",
    )

    return kde_matrix



"""
This module uses data from FRED to find values for parameters for the
OG-USA model that rely on macro data for calibration.
"""

# imports
import pandas_datareader.data as web
import pandas as pd
import numpy as np
import datetime
import statsmodels.api as sm


def get_macro_params():
    """
    Compute values of parameters that are derived from macro data
    """

    # set beginning and end dates for data
    # format is year (1940),month (1),day (1)
    start = datetime.datetime(1947, 1, 1)
    end = datetime.date.today()  # go through today
    baseline_date = datetime.datetime(2023, 3, 31)

    variable_dict = {
        "GDP Per Capita": "A939RX0Q048SBEA",
        "Labor share": "LABSHPUSA156NRUG",
        "Debt held by foreigners": "FDHBFIN",
        "Debt held by public": "FYGFDPUN",
        "BAA Corp Bond Rates": "DBAA",
        "10 year treasury rate": "DGS10",
        "Total gov transfer payments": "B087RC1Q027SBEA",
        "Social Security payments": "W823RC1",
        "Gov expenditures": "FGEXPND",
        "Gov investment": "A782RC1Q027SBEA",
        "Gov interest payments": "A091RC1Q027SBEA",
        "Real GDP": "GDPC1",
        "Nominal GDP": "GDP",
    }

    # pull series of interest using pandas_datareader
    fred_data = web.DataReader(variable_dict.values(), "fred", start, end)
    fred_data.rename(
        columns=dict((y, x) for x, y in variable_dict.items()), inplace=True
    )

    # make sure all dollar value data are in billions
    fred_data["Debt held by public"] = fred_data["Debt held by public"] / 1000

    # Separate quartely, monthly, and annual dataseries
    fred_data_q = (
        fred_data[
            [
                "Debt held by public",
                "Nominal GDP",
                "Total gov transfer payments",
                "Social Security payments",
                "Gov expenditures",
                "Gov investment",
                "Gov interest payments",
                "GDP Per Capita",
            ]
        ]
        .resample("Q")
        .mean()
    )
    fred_data_a = fred_data[["Labor share"]]
    fred_data_d = fred_data[["BAA Corp Bond Rates", "10 year treasury rate"]]

    # initialize a dictionary of parameters
    macro_parameters = {}

    # print(fred_data.loc(str(baseline_date)))
    # find initial_debt_ratio
    macro_parameters["initial_debt_ratio"] = pd.Series(
        fred_data_q["Debt held by public"] / fred_data_q["Nominal GDP"]
    ).loc[baseline_date]

    # find alpha_T
    macro_parameters["alpha_T"] = pd.Series(
        (
            fred_data_q["Total gov transfer payments"]
            - fred_data_q["Social Security payments"]
        )
        / fred_data_q["Nominal GDP"]
    ).loc[baseline_date]

    # find alpha_G
    macro_parameters["alpha_G"] = pd.Series(
        (
            fred_data_q["Gov expenditures"]
            - fred_data_q["Total gov transfer payments"]
            - fred_data_q["Gov interest payments"]
            - fred_data_q["Gov investment"]
            - fred_data_q["Social Security payments"]
        )
        / fred_data_q["Nominal GDP"]
    ).loc[baseline_date]

    # find alpha_I
    macro_parameters["alpha_I"] = pd.Series(
        fred_data_q["Gov investment"] / fred_data_q["Nominal GDP"]
    ).loc[baseline_date]

    # find gamma
    macro_parameters["gamma"] = 1 - fred_data_a["Labor share"].mean()

    # find g_y
    macro_parameters["g_y"] = (
        fred_data_q["GDP Per Capita"].pct_change(periods=4, freq="Q").mean()
    )

    # # estimate r_gov_shift and r_gov_scale
    rate_data = (
        fred_data_d[["10 year treasury rate", "BAA Corp Bond Rates"]].dropna()[
            -50:
        ]
    ) / 100  # divide by 100 bc data in percentage points
    rate_data["constant"] = np.ones(len(rate_data.index))
    mod = sm.OLS(
        rate_data["10 year treasury rate"],
        rate_data[["constant", "BAA Corp Bond Rates"]],
    )
    res = mod.fit()
    macro_parameters["r_gov_scale"] = res.params["BAA Corp Bond Rates"]
    macro_parameters["r_gov_shift"] = res.params["constant"] * -1

    return macro_parameters



import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import scipy.optimize as opt
from scipy.stats import kde
import os


def MVKDE(
    S, J, proportion_matrix=None, filename=None, plot=False, bandwidth=0.25
):
    """
    Generates a Multivariate Kernel Density Estimator and returns a matrix
    representing a probability distribution according to given age categories,
    and ability type categories.
    Inputs:
        S   	 			  = scalar, the number of age groups in the model.
        J                     = scalar, the number of ability type groups in the model.
        proportion_matrix     = [S, J], array containing the proportion matrix created
                                by SCFExtract.py. This argument would be used if you are
                                passing in the proportion_matrix directly
        filename     		  = string, the file name of the .txt document that contains
                                                        the original proportion matrix created by SCFExtract.py.
                                Use this argument if you have saved the proportion matrix
                                in a .txt file
        plot                  = boolean, whether or not you want a plot of the probability
                                                        distribution generated by your given age and ability type
                                                        groups.
    Functions called:
        kde.gaussian_kde      = scipy function that generates a Kernel
                                                        Density Estimator from given data
    Objects in function:
        proportion_matrix     = [78, 7], array containing the proportion (0 < x < 1) of
                                the total bequests that each age-income category receives.
                                Derived in SCFExtract.py
        age_probs             = [78,], array containing the frequency, or how many times,
                                that random drawn numbers fell into the 78 different age bins
        income_probs          = [7,], array containing the frequency, or how many times,
                                that random drawn numbers fell into the 7 different ability
                                type bins
        age_frequency         = [70000,], array containing repeated age values for each
                                age number at the frequency given by the age_probs vector
        xmesh                 = complex number, the number of age values that will be
                                evaluated in the Kernel Density Estimator.
        freq_mat              = [70000, 2], array containing age_frequency and
                                income_frequency stacked
        density               = object, class given by scipy.stats.gaussian_kde.
                                The Multivariate Kernel Density Estimator for the given data set.
        age_min, age_max      = scalars, the minimum and maximum age values and minimum
        income_min, income_max  and maximum income values
        agei  				  = [S, J], array containing the age values to be evaluated in
                                the Kernel Estimator (ranging from 18-90)
        incomei 			  = [S, J], array containing the income values to be evaluated
                                in the Kernel Estimator (ranging from 1-7)
        coords                = [2, S*J], array containing the raveled values of agei
                                and incomei stacked
        estimator             = [S, J], array containing the new proportion values for
                                s age groups and e ability type groups that are evaluated
                                using the Multivariate Kernel Density Estimator
        estimator_scaled       = [S, J], array containing the new proportion values for
                                s age groups and e ability type groups that are evaluated using
                                the Multivariate Kernel Density Estimator, but scaled so that
                                the sum of the array is equal to one.
    Returns: estimator_scaled
    """
    if proportion_matrix is None:
        proportion_matrix = np.loadtxt(filename, delimiter=",")
    proportion_matrix_income = np.sum(proportion_matrix, axis=0)
    proportion_matrix_age = np.sum(proportion_matrix, axis=1)
    age_probs = np.random.multinomial(70000, proportion_matrix_age)
    income_probs = np.random.multinomial(70000, proportion_matrix_income)
    age_frequency = np.array([])
    income_frequency = np.array([])
    age_mesh = complex(str(S) + "j")
    income_mesh = complex(str(J) + "j")

    j = 18
    """creating a distribution of age values"""
    for i in age_probs:
        listit = np.ones(i)
        listit *= j
        age_frequency = np.append(age_frequency, listit)
        j += 1

    k = 1
    """creating a distribution of ability type values"""
    for i in income_probs:
        listit2 = np.ones(i)
        listit2 *= k
        income_frequency = np.append(income_frequency, listit2)
        k += 1

    freq_mat = np.vstack((age_frequency, income_frequency)).T
    density = kde.gaussian_kde(freq_mat.T, bw_method=bandwidth)
    age_min, income_min = freq_mat.min(axis=0)
    age_max, income_max = freq_mat.max(axis=0)
    agei, incomei = np.mgrid[
        age_min:age_max:age_mesh, income_min:income_max:income_mesh
    ]
    coords = np.vstack([item.ravel() for item in [agei, incomei]])
    estimator = density(coords).reshape(agei.shape)
    estimator_scaled = estimator / float(np.sum(estimator))
    if plot == True:
        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")
        ax.plot_surface(agei, incomei, estimator_scaled, rstride=5)
        ax.set_xlabel("Age")
        ax.set_ylabel("Ability Types")
        ax.set_zlabel("Received proportion of total bequests")
        plt.show()
    return estimator_scaled


def wealth_dist(year_start, year_end, S, J, path):
    """
    FROM THE SURVEY OF CONSUMER FINANCES
    Variables used:
    X8022           FOR THE RESPONDENT, THIS VARIABLE CONTAINS THE DATE-OF-BIRTH (age)
    Net Worth		Already adjusted for inflation, respondent's perception of own
                                    net worth, summary variable
    wgt             Summary variable, weights provided by the Federal Reserve
                    to adjust for selection bias in the survey.
    networth / Ability Type categories:
    1.  Under $15000
    2.  15,000-24,999
    3.  25,000-49,999
    4.  50,000-74,999
    5.  75,000-99,999
    6.  100,000-249,999
    7.  250,000 or more
    Age categories:
    (min age in SCF generally is 18)
    ages 18-96
    (max in 2013 was 95 yrs old)
    Inputs:
    year_start   	 	  = scalar, the first year of the range of data years
                                                    that you want from the survey.
    year_end              = scalar, the end year of the range of data years
                                                    that you want from the survey
    S   	 			  = scalar, the number of age groups in the model.
    J                     = scalar, the number of ability type groups in the model.
    path                  = string, the location of your SCF data files. Instructions on
                            how to download these files is located in HowToFindData.txt
                            You can use dir = os.path.dirname(__file__) + '/' which gives you
                            the path to the directory in which you currently are.
    Functions called:
    initSCFvars	      	  = creates a dataframe with the relevant age, networth, and
                                                    weights variables
    genNetworthMatrix     = creates a matrix full of the networth values for each
                                                    age and income group
    MVKDE                 = generates a Multivariate Kernel Density Estimator according
                            to the data given
    Returns: kernel
    """

    """Loading in data sets into different panda dataframes"""

    filenames = [
        "p13i6",
        "p10i6",
        "p07i6",
        "p04i6",
        "p01i6",
        "p98i6",
        "p95i6",
        "p92i4",
        "p89i6",
    ]

    summaryfiles = [
        "rscfp2013",
        "rscfp2010",
        "rscfp2007",
        "rscfp2004",
        "rscfp2001",
        "rscfp1998",
        "rscfp1995",
        "rscfp1992",
        "rscfp1989",
    ]

    """Making sure feasible inputs are given"""
    if (year_start > 2013) or (year_start < 1989) or (year_start % 3 != 0):
        raise ValueError("SCF data non-existent for start year provided")
    if (year_end > 2013) or (year_end < 1989) or (year_end % 3 != 0):
        raise ValueError("SCF data non-existent for end year provided")
    if (S < 0) or (J < 0):
        raise ValueError("Values must be positive for age and ability types")

    """initializing the year range to be used in SCF data extraction"""
    year_range = np.arange(year_end, year_start - 3, -1)
    year_list = []
    for year in year_range:
        if year % 3 == 0:
            year_list.append(year)
        else:
            pass

    """intializing dataframe lists with SCF and SCF summary variable filenames"""
    df_list = []
    for i in range(len(year_list)):
        df_list.append(pd.read_stata(path + filenames[i] + ".dta"))
    dfs_list = []
    for i in range(len(year_list)):
        dfs_list.append(pd.read_stata(path + summaryfiles[i] + ".dta"))

    """set different income levels for the different ability types"""
    income_levels = [
        -999999999,
        15000,
        25000,
        50000,
        75000,
        100000,
        250000,
        999999999,
    ]

    """variable for the y axis income groups for graph"""
    income_y = np.arange(1, 8)

    """the minimum age in the survey"""
    min_age = 18

    """the maximum age that we want to represent in the data"""
    max_age = 95

    """creating an array full of the different ages"""
    age_groups = np.arange(min_age, max_age + 1)
    income_groups = 7

    def initSCFvars(scf, scfSummary, year, year_range):
        """
        Generates variables from the SCF that will eventually be used
        to calculate total bequests from all respondents, total
        bequests for age-income categories, and a proportion matrix.
        Inputs:
            scf   	 			  = dictionary with all SCF data, indexed with codebook
                                                    variables.
            scfSummary            = dictionary with all SCF summary variables adjusted for inflation.
            year     			  = scalar, what year the SCF survey was taken
            year_range            = list, years to be adjusted for inflation
        Functions called: None
        Objects in function:
            age                  = panda dataframe, contains the ages of the respondents
            age_income_matrix     = [X = # of income groups, Y = # of age groups],
                                                        Nested lists, this variable contains a list of Y lists
                                                        corresponding to each age group, and each of the Y
                                                        lists contains X amounts of income groups. These nested lists
                                                        will be filled with pandas and will be used to extract inheritance
                                                        data corresponding to the age and income conditions.
            income           	  = panda dataframe, contains total household income for each respondent
            age_income 			  = panda dataframe, combines the age, income, value_of_transfer,
                                                        and year_received pandas into one panda.
            total_networth  	  = scalar, total networth of all respondents
            weights               = dataframe, contains the weights corresponding to each respondent,
                                    these correct for selection bias in survey.
        Returns: age_income, total_bequests_received, age_income_matrix, total_bequest_matrix
        """
        age = scf["X8022"]
        weights = scfSummary["wgt"]
        income = scfSummary["networth"] * (weights / 5.0)
        age_income = pd.concat([age, income], axis=1)
        age_income.columns = ["age", "income"]
        total_networth = income.sum()
        return (age_income, total_networth)

    def genNetworthMatrix(
        age_income, age_groups, income_groups, min_age, income_levels, year
    ):
        """
        Generates a matrix for the total bequests of each age-income category,
        after generating a Nested list containing the pandas corresponding to the age,
        income and recent wealth transfer conditions.
        Inputs:
            age_groups            = [S, ], vector of integers corresponding to different ages
                                                        of respondents.
            income_groups         = scalar, total number of income groups/ ability types.
            min_age               = scalar, minimum age of respondents
            income_levels         = [J, ], vector of scalars, corresponding to different
                                                        income cut-offs.
            age_income            = panda dataframe, age, and income columns
            year                  = scalar, what year the SCF survey was taken
        Functions called: None
        Objects in function:
            age_income_matrix     = [J = # of income groups, S = # of age groups],
                                                        Nested lists, this variable contains a list of Y lists
                                                        corresponding to each age group, and each of the Y
                                                        lists contains X amounts of income groups. Filled with pandas
                                                        corresponding to the age, income and recent wealth transfer conditions.
            total_networth_matrix = [S = # of age groups, J = # of income_groups], array filled with
                                                        the average networth of each age-income category
        Returns: total_bequest_matrix
        """
        age_income_matrix = [
            [[] for z in range(income_groups)] for x in range(len(age_groups))
        ]
        total_networth_mat = np.zeros((len(age_groups), income_groups))
        for i in age_groups:
            for j in range(income_groups):
                age_income_matrix[i - min_age][j] = age_income[
                    (age_income["age"] == i)
                    & (age_income["income"] < income_levels[j + 1])
                    & (age_income["income"] >= income_levels[j])
                ]
                total_networth_mat[(i - min_age), j] = (
                    age_income_matrix[i - min_age][j]["income"].sum()
                ) / (age_income_matrix[i - min_age][j]["income"]).count()
        return total_networth_mat

    """initialize lists to contain the matrices produced for the different years"""
    proportion_matrix_list = []
    total_networth_list = []
    total_networth_matrix_list = []
    counter = 0

    """creating the different matrices for the different years and inserting them into their respective lists"""
    for year in year_list:
        age_income, total_networth = initSCFvars(
            df_list[counter], dfs_list[counter], year, year_range
        )
        total_networth_mat = genNetworthMatrix(
            age_income, age_groups, income_groups, min_age, income_levels, year
        )
        total_networth_list.append(total_networth)
        total_networth_matrix_list.append(total_networth_mat)
        counter += 1

    """creating the matrix of all years combined"""
    summed_total_networth = np.zeros((len(age_groups), income_groups))
    for i in range(len(year_list)):
        summed_total_networth += total_networth_matrix_list[i]
    all_years_networth = summed_total_networth / float(len(year_list))
    all_years_proportion = all_years_networth / all_years_networth.sum()
    """fitting a kernel to the data, then scaling it into our initial wealth matrix"""
    kernel = MVKDE(S, J, all_years_proportion, None, False)
    kernel = kernel * all_years_networth.sum()
    return kernel


def Kbar_Eq(scale, Kbar, omega0, lambda0, wealth_dist):
    """
    Used in the fsolve as our equation for aggregate capital
    stock.
    Inputs:
            scale       =  scalar, scaling parameter we are estimating
                                      that will be used to scale data to the model's
                                      aggregate capital stock
            Kbar        =  scalar, aggregate initial capital stock
            omega0      =  [S,] array, containing our population dynamics
            lambda0     =  [J,] array, conatining our income dynamics
            wealth_dist =  [J, S] array, contains our SCF distribution of
                                            wealth among different age and networth categories
                                            calculated using our SCF_wealth_dist function
    Objects in Function:
            k_calc      =  scalar, the sum of each age-income category's
                                       average networth multiplied by their population
                                       and income dynamics
    Returns:
            Kbar - k_calc = should be equal to zero since Kbar = k_calc
    """
    S = len(omega0)
    J = len(lambda0)
    k_calc = 0
    for s in range(S):
        for j in range(J):
            k_calc += omega0[s] * lambda0[j] * scale * wealth_dist[j, s]
    return Kbar - k_calc


def init_wealth(Kbar, omega0, lambda0, wealth_dist, scale_tol):
    """
    Used in the fsolve as our equation for aggregate capital
    stock.
    Inputs:
            Kbar        =  scalar, aggregate initial capital stock
            omega0      =  [S,] array, containing our population dynamics
            lambda0     =  [J,] array, conatining our income dynamics
            wealth_dist =  [J, S] array, contains our SCF distribution of
                                            wealth among different age and networth categories
                                            calculated using our SCF_wealth_dist function
            scale_tol   =  scalar, the tolerance to be used in the fsolve
    Objects in Function:
            scale       =  scalar, the factor that we multipliy our SCF
                                       wealth distribution by to get the initial
                                       wealth distribution in our model
    Returns:
            wealth_init = [J,S] array, our initial distribution of wealth_dist
                                      to be used in our model
            scale       =  scalar, the factor that we multipliy our SCF
                                       wealth distribution by to get the initial
                                       wealth distribution in our model
    """
    S = len(omega0)
    J = len(lambda0)
    scale_guess = 0.5
    scale = opt.fsolve(
        Kbar_Eq,
        scale_guess,
        args=(Kbar, omega0, lambda0, wealth_dist),
        xtol=scale_tol,
    )
    wealth_init = scale * wealth_dist
    k_calc = 0
    for s in range(S):
        for j in range(J):
            k_calc += omega0[s] * lambda0[j] * scale * wealth_dist[j, s]
    return wealth_init, scale


def age_income_plot(networth_range, age_range, proportion_matrix, year):
    X, Y = np.meshgrid(networth_range, age_range)
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    ax.plot_surface(X, Y, proportion_matrix, rstride=5)
    ax.set_xlabel("Ability Types")
    ax.set_ylabel("Age")
    ax.set_zlabel("Total networth for years {}".format(str(year)))
    plt.show()



import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt
from linearmodels import PanelOLS
import ogcore  # import just for MPL style file

# Create directory if output directory does not already exist
cur_path = os.path.split(os.path.abspath(__file__))[0]
output_fldr = "csv_output_files"
output_dir = os.path.join(cur_path, "..", "data", output_fldr)
if not os.access(output_dir, os.F_OK):
    os.makedirs(output_dir)


def estimate_profiles(graphs=False):
    """
    Function to estimate deterministic lifecycle profiles of hourly
    earnings.  Follows methodology of Fullerton and Rogers (1993).

    Args:
        graphs (bool): whether to create graphs of profiles

    Returns:
        reg_results (Pandas DataFrame): regression model coefficients
            for lifetime earnings profiles

    """
    # Read in dataframe of PSID data
    df = ogcore.utils.safe_read_pickle(
        os.path.join(
            cur_path, "..", "data", "PSID", "psid_lifetime_income.pkl"
        )
    )

    model_results = {
        "Names": [
            "Constant",
            "",
            "Head Age",
            "",
            "Head Age^2",
            "",
            "Head Age^3",
            "",
            "R-Squared",
            "Observations",
        ]
    }
    cats_pct = ["0-25", "26-50", "51-70", "71-80", "81-90", "91-99", "100"]
    long_model_results = {
        "Lifetime Income Group": [],
        "Constant": [],
        "Age": [],
        "Age^2": [],
        "Age^3": [],
        "Observations": [],
    }
    for i, group in enumerate(cats_pct):
        data = df[df[group] == 1].copy()
        data["ones"] = np.ones(len(data.index))
        mod = PanelOLS(
            data.ln_earn_rate, data[["ones", "age", "age2", "age3"]]
        )
        res = mod.fit(cov_type="clustered", cluster_entity=True)
        # print('Summary for lifetime income group ', group)
        # print(res.summary)
        # Save model results to dictionary
        model_results[group] = [
            res.params["ones"],
            res.std_errors["ones"],
            res.params["age"],
            res.std_errors["age"],
            res.params["age2"],
            res.std_errors["age2"],
            res.params["age3"],
            res.std_errors["age3"],
            res.rsquared,
            res.nobs,
        ]
        long_model_results["Lifetime Income Group"].extend([cats_pct[i], ""])
        long_model_results["Constant"].extend(
            [res.params["ones"], res.std_errors["ones"]]
        )
        long_model_results["Age"].extend(
            [res.params["age"], res.std_errors["age"]]
        )
        long_model_results["Age^2"].extend(
            [res.params["age2"], res.std_errors["age2"]]
        )
        long_model_results["Age^3"].extend(
            [res.params["age3"], res.std_errors["age3"]]
        )
        long_model_results["Observations"].extend([res.nobs, ""])

    reg_results = pd.DataFrame.from_dict(model_results)
    reg_results.to_csv(
        os.path.join(output_dir, "DeterministicProfileRegResults.csv")
    )
    long_reg_results = pd.DataFrame.from_dict(model_results)
    long_reg_results.to_csv(
        os.path.join(output_dir, "DeterministicProfileRegResults_long.csv")
    )

    if graphs:
        # Plot lifecycles of hourly earnings from processes estimated above
        age_vec = np.arange(20, 81, step=1)
        for i, group in enumerate(cats_pct):
            earn_profile = (
                model_results[group][0]
                + model_results[group][2] * age_vec
                + model_results[group][4] * age_vec**2
                + model_results[group][6] * age_vec**3
            )
            plt.plot(age_vec, earn_profile, label=group)
        plt.title(
            "Estimated Lifecycle Earnings Profiles by Lifetime Income Group"
        )
        plt.legend()

        plt.savefig(
            os.path.join(output_dir, "lifecycle_earnings_profiles.png")
        )

        # Plot of lifecycles of hourly earnings from processes from data
        pd.pivot_table(
            df,
            values="ln_earn_rate",
            index="age",
            columns="li_group",
            aggfunc="mean",
        ).plot(legend=True)
        plt.title(
            "Empirical Lifecycle Earnings Profiles by Lifetime Income Group"
        )

        plt.savefig(
            os.path.join(output_dir, "lifecycle_earnings_profiles_data.png")
        )

        # Plot of lifecycle profiles of hours by lifetime income group
        # create variable from fraction of time endowment work
        df["labor_supply"] = df["earnhours_hh"] / (
            24 * 5 * (df["married"] + 1) * 50
        )
        pd.pivot_table(
            df,
            values="labor_supply",
            index="age",
            columns="li_group",
            aggfunc="mean",
        ).plot(legend=True)
        plt.title("Lifecycle Profiles of Hours by Lifetime Income Group")

        plt.savefig(os.path.join(output_dir, "lifecycle_laborsupply.png"))

    return reg_results
